import numpy as np
import dcmri.pk as pk


def params_kidney(kinetics) -> list:
    """Kidney model parameters

    Args:
        kinetics (str): kidney model. Options are: 2CF.

    Returns:
        list: Parameter short names
    """
    if kinetics == '2CF':
        return ['Fp', 'vp', 'FF', 'Tt']
    elif kinetics == 'HF':
        return ['vp', 'Ft', 'Tt']
    

def conc_kidney(ca: np.ndarray, *params, t=None, dt=1.0, sum=True, kinetics='2CF', **kwargs) -> np.ndarray:
    """Concentration in kidney tissues.

    Args:
        ca (array-like): concentration in the arterial input.
        params (tuple): free model parameters.
        t (array_like, optional): the time points in sec of the input function *ca*. If *t* is not provided, the time points are assumed to be uniformly spaced with spacing *dt*. Defaults to None.
        dt (float, optional): spacing in seconds between time points for uniformly spaced time points. This parameter is ignored if *t* is explicity provided. Defaults to 1.0.
        kinetics (str, optional): Kinetics of the tissue, either '2CF', 'FN' - see below for detail. Defaults to '2CF'. 
        sum (bool, optional): For two-compartment tissues, set to True to return the total tissue concentration. Defaults to True.
        kwargs (dict, optional): any optional keyword parameters required by the kinetic model - see below for detail.

    Returns:
        numpy.ndarray: If sum=True, this is a 1D array with the total concentration at each time point. If sum=False this is the concentration in each compartment, and at each time point, as a 2D array with dimensions *(2,k)*, where *k* is the number of time points in *ca*. The concentration is returned in units of M.

    Notes:
        Currently implemented kinetic models are: 

        - '2CF': two-compartment filtration model. params = (Fp, Tp, Ft, Tt,)
        - 'FN': free nephron model. params = (Fp, Tp, Ft, h, ). 

        The model parameters are:

        - **Fp** (float, mL/sec/mL): Plasma flow.
        - **Tp** (float, sec): plasma mean transit time.
        - **Ft** (float, mL/sec/mL): tubular flow.
        - **Tt** (float, sec): tubular mean transit time.
        - **hh** (array-like, 1/sec): frequences of transit time histogram. The boundaries of the transit time bins can be provided as an array in a keyword parameter TT, which has to have one more element than h. If TT is not provided, the transit time bins are equally space in the range [0, tmax], where tmax is the largest acquisition time.


    Example:

        Plot concentration in cortex and medulla for typical values:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import dcmri as dc

        Generate a population-average input function:

        >>> t = np.arange(0, 300, 1.5)
        >>> ca = dc.aif_parker(t, BAT=20)

        Define some parameters and generate plasma and tubular tissue concentrations with a 2-compartment filtration model:

        >>> Fp, Tp, Ft, Tt = 0.05, 10, 0.01, 120
        >>> C = dc.conc_kidney(ca, Fp, Tp, Ft, Tt, t=t, sum=False, kinetics='2CF')

        Plot all concentrations:

        >>> fig, ax = plt.subplots(1,1,figsize=(6,5))
        >>> ax.set_title('Kidney concentrations')
        >>> ax.plot(t/60, 1000*C[0,:], linestyle='--', linewidth=3.0, color='darkred', label='Plasma')
        >>> ax.plot(t/60, 1000*C[1,:], linestyle='--', linewidth=3.0, color='darkblue', label='Tubuli')
        >>> ax.plot(t/60, 1000*(C[0,:]+C[1,:]), linestyle='-', linewidth=3.0, color='grey', label='Whole kidney')
        >>> ax.set_xlabel('Time (min)')
        >>> ax.set_ylabel('Tissue concentration (mM)')
        >>> ax.legend()
        >>> plt.show()

        Use generate plasma and tubular tissue concentrations using the free nephron model for comparison. We assume 4 transit time bins with the following boundaries (in units of seconds):

        >>> TT = [0, 15, 30, 60, 120]

        with longest transit times most likely (note the frequences to not have to add up to 1):

        >>> h = [1, 2, 3, 4]
        >>> C = dc.conc_kidney(ca, Fp, Tp, Ft, h, t=t, sum=False, kinetics='FN', TT=TT)

        Plot all concentrations:

        >>> fig, ax = plt.subplots(1,1,figsize=(6,5))
        >>> ax.set_title('Kidney concentrations')
        >>> ax.plot(t/60, 1000*C[0,:], linestyle='--', linewidth=3.0, color='darkred', label='Plasma')
        >>> ax.plot(t/60, 1000*C[1,:], linestyle='--', linewidth=3.0, color='darkblue', label='Tubuli')
        >>> ax.plot(t/60, 1000*(C[0,:]+C[1,:]), linestyle='-', linewidth=3.0, color='grey', label='Whole kidney')
        >>> ax.set_xlabel('Time (min)')
        >>> ax.set_ylabel('Tissue concentration (mM)')
        >>> ax.legend()
        >>> plt.show()

    """
    if kinetics == '2CF':
        return _conc_kidney_2cf(ca, *params, t=t, dt=dt, sum=sum)
    elif kinetics == 'HF':
        return _conc_kidney_hf(ca, *params, t=t, dt=dt, sum=sum)
    elif kinetics == 'FN':
        # TT = [15,30,60,90,150,300,600]
        return _conc_kidney_fn(ca, *params, t=t, dt=dt, sum=sum, **kwargs)
    else:
        raise ValueError(
            'Kinetic model ' + kinetics + ' is not currently implemented.')


def _conc_kidney_2cf(ca, Fp, vp, Ft, Tt, t=None, dt=1.0, sum=True):
    #vp = Tp*(Fp+Ft)
    Tp = vp/(Fp+Ft)
    Cp = pk.conc_comp(Fp*ca, Tp, t=t, dt=dt)
    cp = Cp/vp
    Ct = pk.conc_comp(Ft*cp, Tt, t=t, dt=dt)
    if sum:
        return Cp+Ct
    else:
        return np.stack((Cp, Ct))


def _conc_kidney_hf(ca, vp, Ft, Tt, t=None, dt=1.0, sum=True):
    Cp = vp*ca
    Ct = pk.conc_comp(Ft*ca, Tt, t=t, dt=dt)
    if sum:
        return Cp+Ct
    else:
        return np.stack((Cp, Ct))


def _conc_kidney_fn(ca, Fp, Tp, Ft, h, t=None, dt=1.0, sum=True, TT=None):
    if TT is None:
        if t is None:
            tmax = dt*np.size(ca)
        else:
            tmax = np.amax(t)
        nTT = 1+np.size(h)
        TT = np.linspace(0, tmax, nTT)
    vp = Tp*(Fp+Ft)
    Cp = pk.conc_plug(Fp*ca, Tp, t=t, dt=dt)
    cp = Cp/vp
    Ct = pk.conc_free(Ft*cp, h, dt=dt, TT=TT, solver='step')
    if sum:
        return Cp+Ct
    else:
        return np.stack((Cp, Ct))


def conc_kidney_cm(ca: np.ndarray, *params, t=None, dt=1.0, sum=True, 
                   kinetics='7C'):
    """Concentration in kidney cortex and medulla tissues.

    Args:
        ca (array-like): concentration in the arterial input.
        params (tuple): free model parameters.
        t (array_like, optional): the time points in sec of the input 
          function *ca*. If *t* is not provided, the time points are assumed 
          to be uniformly spaced with spacing *dt*. Defaults to None.
        dt (float, optional): spacing in seconds between time points for 
          uniformly spaced time points. This parameter is ignored if *t* is 
          explicity provided. Defaults to 1.0.
        sum (bool, optional): For two-compartment tissues, set to True to 
          return the total tissue concentration. Defaults to True.
        kinetics (str, optional): Kinetics of the tissue, currently only '7C' 
          available - see below for detail. Defaults to '7F'. 

    Returns:
        tuple[numpy.ndarray, numpy.ndarray]: If sum=True, each return value 
        is a 1D array with the total concentration at each time point, in 
        cortex and medulla, respectively. If sum=False each return value is 
        the concentration in each compartment, and at each time point, of 
        cortex and medulla as a 2D array with dimensions *(n,k)*, where n is 
        the number of compartments and *k* is the number of time points in 
        *ca*. The concentration is returned in units of M.


    Notes:
        Currently implemented kinetic models are: 

        - '7CF': 7-compartment model. 
          params = (Fp, Eg, fc, Tg, Tv, Tpt, Tlh, Tdt, Tcd,). 
          Cortico-medullary model with 4 cortical compartments (glomeruli, 
          peritubular capillaries & veins, proximal tubuli and distal tubuli) 
          and 3 medullary compartments (peritubular capillaries & veins, 
          list of Henle and collecting ducts). 

        The 9 model parameters are:

        - **Fp** (float): Plasma flow into the tissue, in units of mL 
          plasma per sec and per mL tissue (mL/sec/mL).
        - **Eg** (float): Glomerular extraction fraction
        - **fc** (float): Cortical flow fraction
        - **Tg** (float): Glomerular mean transit time in sec
        - **Tv** (float): Peritubular & venous mean transit time in sec
        - **Tpt** (float): Proximal tubuli mean transit time in sec
        - **Tlh** (float): Lis of Henle mean transit time in sec
        - **Tdt** (float): Distal tubuli mean transit time in sec
        - **Tcd** (float): Collecting duct mean transit time in sec

    Example:

        Plot concentration in cortex and medulla for typical values:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import dcmri as dc

        Generate a population-average input function:

        >>> t = np.arange(0, 300, 1.5)
        >>> ca = dc.aif_parker(t, BAT=20)

        Use the function to generate total cortex and medulla tissue concentrations:

        >>> Fp, Eg, fc, Tg, Tv, Tpt, Tlh, Tdt, Tcd = 0.03, 0.15, 0.8, 4, 10, 60, 60, 30, 30
        >>> Cc, Cm = dc.conc_kidney_cm(ca, Fp, Eg, fc, Tg, Tv, Tpt, Tlh, Tdt, Tcd, t=t, kinetics='7C')

        Plot all concentrations:

        >>> fig, ax = plt.subplots(1,1,figsize=(6,5))
        >>> ax.set_title('Kidney concentrations')
        >>> ax.plot(t/60, 1000*Cc, linestyle='-', linewidth=3.0, color='darkblue', label='Cortex')
        >>> ax.plot(t/60, 1000*Cm, linestyle='-', linewidth=3.0, color='darkgreen', label='Medulla')
        >>> ax.plot(t/60, 1000*(Cc+Cm), linestyle='-', linewidth=3.0, color='darkgrey', label='Whole kidney')
        >>> ax.set_xlabel('Time (min)')
        >>> ax.set_ylabel('Tissue concentration (mM)')
        >>> ax.legend()
        >>> plt.show()
    """
    if kinetics == '7C':
        return _conc_kidney_cm9(ca, *params, t=t, dt=dt, sum=sum)
    else:
        raise ValueError(
            'Kinetic model ' + kinetics + ' is not currently implemented.')


def _conc_kidney_cm9(ca, Fp, Eg, fc, Tg, Tv, Tpt, Tlh, Tdt, Tcd, t=None, dt=1.0, sum=True):

    # Flux out of the glomeruli and arterial tree
    Jg = pk.flux(Fp*ca, Tg, t=t, dt=dt, model='comp')

    # Flux out of the peritubular capillaries and venous system
    Jv = pk.flux((1-Eg)*Jg, Tv, t=t, dt=dt, model='comp')

    # Flux out of the proximal tubuli
    Jpt = pk.flux(Eg*Jg, Tpt, t=t, dt=dt, model='comp')

    # Flux out of the lis of Henle
    Jlh = pk.flux(Jpt, Tlh, t=t, dt=dt, model='comp')

    # Flux out of the distal tubuli
    Jdt = pk.flux(Jlh, Tdt, t=t, dt=dt, model='comp')

    # Flux out of the collecting ducts
    Jcd = pk.flux(Jdt, Tcd, t=t, dt=dt, model='comp')

    # Build cortical concentrations
    Cg = Tg*Jg      # arteries/glomeruli
    Cv = fc*Tv*Jv   # part of the peritubular capillaries
    Cpt = Tpt*Jpt   # proximal tubuli
    Cdt = Tdt*Jdt   # distal tubuli
    if sum:
        Ccor = Cg + Cv + Cpt + Cdt
    else:
        Ccor = np.stack((Cg, Cv, Cpt, Cdt))

    # Build medullary concentrations
    Cv = (1-fc)*Tv*Jv   # part of the peritubular capillaries
    Clh = Tlh*Jlh       # Lis of Henle
    Ccd = Tcd*Jcd       # collecting ducts
    if sum:
        Cmed = Cv + Clh + Ccd
    else:
        Cmed = np.stack((Cv, Clh, Ccd))

    return Ccor, Cmed



import matplotlib.pyplot as plt
import numpy as np

import dcmri.kidney as pkk
import dcmri.lib as lib
import dcmri.ui as ui
import dcmri.sig as sig
import dcmri.utils as utils


class KidneyCortMed(ui.Model):
    """
    General model for renal cortico-medullary data.

    **warning**: This model is functional but under active 
    development. Future versions may change without warning.

        **Kinetic parameters**

        - **Hct** (float, optional): Hematocrit. 
        - **Fp** Plasma flow in mL/sec/mL. 
        - **Eg** Glomerular extraction fraction. 
        - **fc** Cortical flow fraction. 
        - **Tg** Glomerular mean transit time in sec. 
        - **Tv** Peritubular & venous mean transit time in sec. 
        - **Tpt** Proximal tubuli mean transit time in sec. 
        - **Tlh** Lis of Henle mean transit time in sec. 
        - **Tdt** Distal tubuli mean transit time in sec. 
        - **Tcd** Collecting duct mean transit time in sec.

        **Input function**

        - **aif** (array-like, default=None): Signal-time curve in a feeding artery. If AIF is set to None, then the parameter ca must be provided (arterial concentrations).
        - **ca** (array-like, default=None): Concentration (M) in the arterial input. Must be provided when aif = None, ignored otherwise.

        **Acquisition parameters**

        - **sequence** (str, default='SR'): imaging sequence.
        - **dt** (float, optional): Sampling interval of the AIF in sec. 
        - **agent** (str, optional): Contrast agent generic name.
        - **field_strength** (float, optional): Magnetic field strength in T. 
        - **TR** (float, optional): Repetition time, or time between excitation pulses, in sec. 
        - **FA** (float, optional): Nominal flip angle in degrees.
        - **Tsat** (float, optional): Time before start of readout (sec).
        - **TC** (float, optional): Time to the center of the readout pulse
        - **n0** (float, optional): Baseline length in number of acquisitions.

        **Signal parameters**

        - **R10c** (float, optional): Precontrast cortex relaxation rate in 1/sec. 
        - **R10m** (float, optional): Precontrast medulla relaxation rate in 1/sec.
        - **R10a** (float, optional): Precontrast arterial relaxation rate in 1/sec. 
        - **S0c** (float, optional): Signal scaling factor in the cortex (a.u.).
        - **S0m** (float, optional): Signal scaling factor in the medulla (a.u.).

        **Prediction and training parameters**

        - **free** (array-like): list of free parameters. The default depends on the kinetics parameter.
        - **free** (array-like): 2-element list with lower and upper free of the free parameters. The default depends on the kinetics parameter.

        **Other parameters**

        - **vol** (float, optional): Liver volume in mL.


    See Also:
        `Kidney`, `Liver`

    Example:

        Derive model parameters from simulated data:

    .. plot::
        :include-source:
        :context: close-figs

        >>> import dcmri as dc

        Use `fake_kidney` to generate synthetic test data:

        >>> time, aif, roi, gt = dc.fake_kidney(CNR=100)

        Build a tissue model and set the constants to match the experimental conditions of the synthetic test data:

        >>> model = dc.KidneyCortMed(
        ...     aif = aif,
        ...     dt = time[1],
        ...     agent = 'gadoterate',
        ...     TR = 0.005,
        ...     FA = 15,
        ...     TC = 0.2,
        ...     n0 = 10,
        ... )

        Train the model on the ROI data and predict signals and concentrations:

        >>> model.train(time, roi)

        Plot the reconstructed signals (left) and concentrations (right) and compare the concentrations against the noise-free ground truth:

        >>> model.plot(time, roi, ref=gt)
    """

    free = {}   #: lower- and upper free for all free parameters.

    def __init__(self, sequence='SR', **params):

        # Config
        self.sequence = sequence

        # Input function
        self.aif = None
        self.ca = None

        # Acquisition parameters
        self.field_strength = 3.0
        self.t = None
        self.dt = 0.5
        self.agent = 'gadoterate'
        self.n0 = 1
        self.TR = 0.005
        self.FA = 15.0
        self.Tsat = 0
        self.TC = 0.085
        self.TS = None

        # Tracer-kinetic parameters
        self.Hct = 0.45
        self.Fp = 0.03
        self.Eg = 0.15
        self.fc = 0.8
        self.Tg = 4
        self.Tv = 10
        self.Tpt = 60
        self.Tlh = 60
        self.Tdt = 30
        self.Tcd = 30

        # Signal parameters
        self.R10c = 1/lib.T1(3.0, 'kidney')
        self.R10m = 1/lib.T1(3.0, 'kidney')
        self.R10a = 1/lib.T1(3.0, 'blood')
        self.S0c = 1
        self.S0m = 1

        # Other parameters
        self.vol = None

        # training parameters
        self.free = {
            'Fp': [0.01, 1],
            'Eg': [0, 1],
            'fc': [0, 1],
            'Tg': [0, 10],
            'Tv': [0, 30],
            'Tpt': [0, np.inf],
            'Tlh': [0, np.inf],
            'Tdt': [0, np.inf],
            'Tcd': [0, np.inf],
        }

        # overide defaults
        for k, v in params.items():
            setattr(self, k, v)

        # Check inputs
        if (self.aif is None) and (self.ca is None):
            raise ValueError(
                'Please provide either an arterial sigal (aif) or an arterial concentration (ca).')

    def time(self):
        """Array of time points

        Returns:
            np.ndarray: time points in seconds.
        """
        if self.t is None:
            if self.aif is None:
                return self.dt*np.arange(np.size(self.ca))
            else:
                return self.dt*np.arange(np.size(self.aif))
        else:
            return self.t

    def conc(self, sum=True):
        """Cortical and medullary concentration

        Returns:
            tuple: Concentration in cortex, concentration in medulla, in M
        """

        if self.aif is not None:
            r1 = lib.relaxivity(self.field_strength, 'blood', self.agent)
            if self.sequence == 'SR':
                cb = sig.conc_src(self.aif, self.TC, 1/self.R10a, r1, self.n0)
            elif self.sequence == 'SS':
                cb = sig.conc_ss(self.aif, self.TR, self.FA,
                                 1/self.R10a, r1, self.n0)
            else:
                raise NotImplementedError(
                    'Signal model ' + self.sequence + 'is not (yet) supported.')
            self.ca = cb/(1-self.Hct)
        return pkk.conc_kidney_cm(self.ca,
                                  self.Fp, self.Eg, self.fc, self.Tg, self.Tv,
                                  self.Tpt, self.Tlh, self.Tdt, self.Tcd,
                                  t=self.t, dt=self.dt, sum=sum, kinetics='7C')

    def predict(self, xdata: np.ndarray) -> tuple:
        """Predict the data at given xdata

        Args:
            xdata (array-like): Either an array with x-values (time points) or a tuple with multiple such arrays

        Returns:
            tuple[np.ndarray,np.ndarray]: tuple of two 1D arrays with signal in corex and medulla, respectively.
        """
        Cc, Cm = self.conc()
        r1 = lib.relaxivity(self.field_strength, 'blood', self.agent)
        R1c = self.R10c + r1*Cc
        R1m = self.R10m + r1*Cm
        if self.sequence == 'SR':
            Sc = sig.signal_spgr(self.S0c, R1c, self.TC, self.TR,
                               self.FA, self.Tsat)
            Sm = sig.signal_spgr(self.S0m, R1m, self.TC, self.TR,
                               self.FA, self.Tsat)
        elif self.sequence == 'SS':
            Sc = sig.signal_ss(self.S0c, R1c, self.TR, self.FA)
            Sm = sig.signal_ss(self.S0m, R1m, self.TR, self.FA)
        t = self.time()
        return (
            utils.sample(xdata, t, Sc, self.TS),
            utils.sample(xdata, t, Sm, self.TS),
        )

    def train(self, xdata: np.ndarray, ydata: tuple[np.ndarray, np.ndarray], **kwargs):
        """Train the free parameters

        Args:
            xdata (array-like): Array with x-data (time points)
            ydata (tuple): Tuple of two 1D arrays (signal_cortex, signal_mdeulla) with signals in cortex and medulla, respectively.
            kwargs: any keyword parameters accepted by `scipy.optimize.curve_fit`.

        Returns:
            Model: A reference to the model instance.
        """

        if self.sequence == 'SR':
            Scref = sig.signal_spgr(1, self.R10c, self.TC, self.TR,
                                  self.FA, self.Tsat)
            Smref = sig.signal_spgr(1, self.R10m, self.TC, self.TR,
                                  self.FA, self.Tsat)
        elif self.sequence == 'SS':
            Scref = sig.signal_ss(1, self.R10c, self.TR, self.FA)
            Smref = sig.signal_ss(1, self.R10m, self.TR, self.FA)
        self.S0c = np.mean(ydata[0][:self.n0]) / Scref
        self.S0m = np.mean(ydata[1][:self.n0]) / Smref
        return ui.train(self, xdata, ydata, **kwargs)

    def export_params(self):
        pars = {}
        pars['Fp'] = ['Plasma flow', self.Fp, 'mL/sec/cm3']
        pars['Eg'] = ['Glomerular extraction fraction', self.Eg, '']
        pars['fc'] = ['Cortical flow fraction', self.fc, '']
        pars['Tg'] = ['Glomerular mean transit time', self.Tg, 'sec']
        pars['Tv'] = ['Peritubular & venous mean transit time', self.Tv, 'sec']
        pars['Tpt'] = ['Proximal tubuli mean transit time', self.Tpt, 'sec']
        pars['Tlh'] = ['Lis of Henle mean transit time', self.Tlh, 'sec']
        pars['Tdt'] = ['Distal tubuli mean transit time', self.Tdt, 'sec']
        pars['Tcd'] = ['Collecting duct mean transit time', self.Tcd, 'sec']
        pars['FF'] = ['Filtration fraction', self.Eg/(1-self.Eg), '']
        pars['Ft'] = ['Tubular flow', self.Fp *
                      self.Eg/(1-self.Eg), 'mL/sec/cm3']
        pars['CBF'] = ['Cortical blood flow',
                       self.Fp/(1-self.Hct), 'mL/sec/cm3']
        pars['MBF'] = ['Medullary blood flow',
                       (1-self.fc)*(1-self.Eg)*self.Fp/(1-self.Hct), 'mL/sec/cm3']
        if self.vol is None:
            return self._add_sdev(pars)
        pars['SKGFR'] = ['Single-kidney glomerular filtration rate',
                         self.vol*self.Fp*self.Eg/(1-self.Eg), 'mL/sec']
        pars['SKBF'] = ['Single-kidney blood flow',
                        self.vol*self.Fp/(1-self.Hct), 'mL/sec']
        pars['SKMBF'] = ['Single-kidney medullary blood flow', self.vol *
                         (1-self.fc)*(1-self.Eg)*self.Fp/(1-self.Hct), 'mL/sec']

        return self._add_sdev(pars)

    def plot(self, xdata: np.ndarray, ydata: tuple[np.ndarray, np.ndarray],
             ref=None, xlim=None, fname=None, show=True):
        """Plot the model fit against data

        Args:
            xdata (array-like): Array with x-data (time points)
            ydata (tuple of arrays): Tuple of two 1D arrays (signal_cortex, signal_mdeulla) with signals in cortex and medulla, respectively.
            xlim (array_like, optional): 2-element array with lower and upper boundaries of the x-axis. Defaults to None.
            ref (tuple, optional): Tuple of optional test data in the form (x,y), where x is an array with x-values and y is an array with y-values. Defaults to None.
            fname (path, optional): Filepath to save the image. If no value is provided, the image is not saved. Defaults to None.
            show (bool, optional): If True, the plot is shown. Defaults to True.
        """

        time = self.time()
        if xlim is None:
            xlim = [np.amin(time), np.amax(time)]
        roic_pred, roim_pred = self.predict(time)
        concc, concm = self.conc()

        fig, (ax0, ax1) = plt.subplots(1, 2, figsize=(12, 5))
        ax0.set_title('Prediction of the MRI signals.')
        ax0.plot(xdata/60, ydata[0], marker='o', linestyle='None',
                 color='cornflowerblue', label='Cortex data')
        ax0.plot(xdata/60, ydata[1], marker='x', linestyle='None',
                 color='cornflowerblue', label='Medulla data')
        ax0.plot(time/60, roic_pred, linestyle='-', linewidth=3.0,
                 color='darkblue', label='Cortex prediction')
        ax0.plot(time/60, roim_pred, linestyle='--', linewidth=3.0,
                 color='darkblue', label='Medulla prediction')
        ax0.set(xlabel='Time (min)', ylabel='MRI signal (a.u.)',
                xlim=np.array(xlim)/60)
        ax0.legend()

        ax1.set_title('Reconstruction of concentrations.')
        if ref is not None:
            ax1.plot(ref['t']/60, 1000*ref['Cc'], marker='o', linestyle='None',
                     color='cornflowerblue', label='Cortex ground truth')
            ax1.plot(ref['t']/60, 1000*ref['Cm'], marker='x', linestyle='None',
                     color='cornflowerblue', label='Medulla ground truth')
            ax1.plot(ref['t']/60, 1000*ref['cp'], marker='o', linestyle='None',
                     color='lightcoral', label='Arterial ground truth')
        ax1.plot(time/60, 1000*concc, linestyle='-', linewidth=3.0,
                 color='darkblue', label='Cortex prediction')
        ax1.plot(time/60, 1000*concm, linestyle='--', linewidth=3.0,
                 color='darkblue', label='Medulla prediction')
        ax1.plot(time/60, 1000*self.ca, linestyle='-', linewidth=3.0,
                 color='darkred', label='Arterial prediction')
        ax1.set(xlabel='Time (min)', ylabel='Concentration (mM)',
                xlim=np.array(xlim)/60)
        ax1.legend()

        if fname is not None:
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()

    def cost(self, xdata: np.ndarray, ydata: tuple[np.ndarray, np.ndarray], metric='NRMS') -> float:
        """Return the goodness-of-fit

        Args:
            xdata (array-like): Array with x-data (time points).
            ydata (tuple): Tuple of two 1D arrays (signal_cortex, signal_mdeulla) with signals in cortex and medulla, respectively.
            metric (str, optional): Which metric to use - options are: 
                **RMS** (Root-mean-square);
                **NRMS** (Normalized root-mean-square); 
                **AIC** (Akaike information criterion); 
                **cAIC** (Corrected Akaike information criterion for small models);
                **BIC** (Baysian information criterion). Defaults to 'NRMS'.

        Returns:
            float: goodness of fit.
        """
        return super().cost(xdata, ydata, metric)



import os
import sys
import pickle
import shutil
import zipfile
import csv
from io import TextIOWrapper

import requests
import numpy as np

# filepaths need to be identified with importlib_resources
# rather than __file__ as the latter does not work at runtime
# when the package is installed via pip install

if sys.version_info < (3, 9):
    # importlib.resources either doesn't exist or lacks the files()
    # function, so use the PyPI version:
    import importlib_resources
else:
    # importlib.resources has files(), so use that:
    import importlib.resources as importlib_resources


# Zenodo DOI of the repository
# DOIs need to be updated when new versions are created
DOI = {
    'MRR': "15285017",      # v0.0.3
    'TRISTAN': "15285027"   # v0.0.1
}

# Datasets available via fetch()
DATASETS = {
    'KRUK': {'doi': DOI['MRR'], 'ext': '.dmr.zip'},
    'tristan_humans_healthy_ciclosporin': {'doi': DOI['TRISTAN'], 'ext': '.dmr.zip'},
    'tristan_humans_healthy_controls_leeds': {'doi': DOI['TRISTAN'], 'ext': '.dmr.zip'},
    'tristan_humans_healthy_controls_sheffield': {'doi': DOI['TRISTAN'], 'ext': '.dmr.zip'},
    'tristan_humans_healthy_metformin': {'doi': DOI['TRISTAN'], 'ext': '.dmr.zip'},
    'tristan_humans_healthy_rifampicin': {'doi': DOI['TRISTAN'], 'ext': '.dmr.zip'},
    'tristan_humans_patients_rifampicin': {'doi': DOI['TRISTAN'], 'ext': '.dmr.zip'},
    'tristan_rats_healthy_multiple_dosing': {'doi': DOI['TRISTAN'], 'ext': '.dmr.zip'},
    'tristan_rats_healthy_reproducibility': {'doi': DOI['TRISTAN'], 'ext': '.dmr.zip'},
    'tristan_rats_healthy_six_drugs': {'doi': DOI['TRISTAN'], 'ext': '.dmr.zip'},
    'minipig_renal_fibrosis': {'doi': None, 'ext': '.dmr.zip'},
}


def fetch(dataset=None, clear_cache=False, download_all=False) -> dict:
    """Fetch a dataset included in dcmri

    Args:
        dataset (str, optional): name of the dataset. See below for options.
        clear_cache (bool, optional): When a dataset is fetched, it is 
          downloaded and then stored in a local cache memory for faster access 
          next time it is fetched. Set clear_cache=True to delete all data 
          in the cache memory. Default is False.
        download_all (bool, optional): By default only the dataset that is 
          fetched is downloaded. Set download_all=True to download all 
          datasets at once. This will cost some time but then offers fast and 
          offline access to all datasets afterwards. This will take up around 
          300 MB of space on your hard drive. Default is False.

    Returns:
        dict: Data as a dictionary.

    Notes:

        The following datasets are currently available:

        `Magnetic resonance renography <https://zenodo.org/records/15284968>`_

            - KRUK

        `TRISTAN Gadoxetate kinetics <https://zenodo.org/records/15285027>`_
        
            - tristan_humans_healthy_rifampicin
            - tristan_humans_healthy_metformin
            - tristan_humans_healthy_ciclosporin
            - tristan_humans_healthy_controls_leeds
            - tristan_humans_healthy_controls_sheffield
            - tristan_rats_healthy_six_drugs
            - tristan_rats_healthy_reproducibility
            - tristan_rats_healthy_multiple_dosing

        Other

            - minipig_renal_fibrosis: Kidney data in a minipig with 
              unilateral ureter stenosis. More detail in future versions..


    Example:

    Fetch the **tristan_humans_healthy_rifampicin** dataset and read it:

    .. plot::
        :include-source:
        :context: close-figs

        >>> import dcmri as dc
        >>> import pydmr
        
        # fetch dmr file
        >>> file = dc.fetch('tristan_humans_healthy_rifampicin')

        # read dmr file
        >>> data = pydmr.read(file)

    """

    if dataset is None:
        v = None
    elif dataset not in DATASETS:
        raise ValueError(
            f'Dataset {dataset} is unknown. Please choose one of '
            f'{DATASETS}'
        )        
    else:
        v = _fetch_dataset(dataset)

    if clear_cache:
        _clear_cache()

    if download_all:
        for d in DATASETS.keys():
            _download(d)

    return v


def _clear_cache():
    """
    Clear the folder where the data downloaded via fetch are saved.

    Note if you clear the cache the data will need to be downloaded again 
    if you need them.
    """

    f = importlib_resources.files('dcmri.datafiles')
    for item in f.iterdir(): 
        if item.is_file(): 
            item.unlink() # Delete the file


def _fetch_dataset(dataset):

    f = importlib_resources.files('dcmri.datafiles')
    datafile = str(f.joinpath(dataset + DATASETS[dataset]['ext']))

    # If this is the first time the data are accessed, download them.
    if not os.path.exists(datafile):
        _download(dataset)

    return datafile



def _download(dataset): # add version keyword
        
    f = importlib_resources.files('dcmri.datafiles')
    datafile = str(f.joinpath(dataset + DATASETS[dataset]['ext']))

    if os.path.exists(datafile):
        return

    # Dataset repository
    version_doi = DATASETS[dataset]['doi']
    if version_doi is None:
        raise ValueError(
            f'Dataset {dataset} is not online and not stored in dcmri/datafiles.'
        )

    # Dataset download link
    file_url = "https://zenodo.org/records/" + version_doi + "/files/" + dataset + DATASETS[dataset]['ext']

    # Make the request and check for connection error
    try:
        file_response = requests.get(file_url) 
    except requests.exceptions.ConnectionError as err:
        raise requests.exceptions.ConnectionError(
            "\n\n"
            "A connection error occurred trying to download the test data \n"
            "from Zenodo. This usually happens if you are offline. The \n"
            "first time a dataset is fetched via dcmri.fetch you need to \n"
            "be online so the data can be downloaded. After the first \n"
            "time they are saved locally so afterwards you can fetch \n"
            "them even if you are offline. \n\n"
            "The detailed error message is here: " + str(err)) 
    
    # Check for other errors
    file_response.raise_for_status()

    # Save the file locally 
    with open(datafile, 'wb') as f:
        f.write(file_response.content)


import os
import multiprocessing
import warnings
import pickle

import numpy as np
from scipy.optimize import curve_fit
from tqdm import tqdm


try:
    num_workers = int(len(os.sched_getaffinity(0)))
except:
    num_workers = int(os.cpu_count())


class ArrayModel():
    # Abstract base class for end-to-end models with pixel-based analysis

    def __init__(self):
        self.shape = None
        self.free = {}

    def _par_values(self):
        return {par: getattr(self, par) for par in self.__dict__ 
                if par not in ['free', 'shape']}


    def params(self, *args, round_to=None):
        """Return the parameter values

        Args:
            args (tuple): parameters to get
            round_to (int, optional): Round to how many digits. If this is 
            not provided, the values are not rounded. Defaults to None.

        Returns:
            list or float: values of parameter values, or a scalar value if 
            only one parameter is required.
        """
        return params(self, *args, round_to=round_to)

    def set_free(self, pop=None, **kwargs):
        """Set the free model parameters.

        Args:
            pop (str or list): a single variable or a list of variables to 
              remove from the list of free parameters. 

        Raises:
            ValueError: if the pop argument contains a parameter that is not 
              in the list of free parameters.
            ValueError: If the parameter is not a model parameter, or bounds 
              are not properly formatted.
        """
        for p in kwargs:
            if not self._params()[p]['pixel_par']:
                raise ValueError(
                    str(p) + ' is not a pixel-based parameter. ' +
                    'Only pixel-based parameters can be free.')
        set_free(self, pop=pop, **kwargs)

    def save(self, file=None, path=None, filename='Model'):
        """Save the current state of the model

        Args:
            file (str, optional): complete path of the file. If this is not 
              provided, a file is constructure from path amd filename 
              variables. Defaults to None.
            path (str, optional): path to store the state if file is not 
              provided. This variable is ignored if file is provided. 
              Defaults to current working directory.
            filename (str, optional): filename to store the state if file is 
              not provided. If no extension is included, the extension '.pkl' 
              is automatically added. This variable is ignored if file is 
              provided. Defaults to 'Model'.

        Returns:
            dict: class instance
        """
        return _save(self, file, path, filename)

    def load(self, file=None, path=None, filename='Model'):
        """Load the saved state of the model

        Args:
            file (str, optional): complete path of the file. If this is not 
              provided, a file is constructure from path amd filename 
              variables. Defaults to None.
            path (str, optional): path to store the state if file is not 
              provided. This variable is ignored if file is provided. 
              Defaults to current working directory.
            filename (str, optional): filename to store the state if file is 
              not provided. If no extension is included, the extension 
              '.pkl' is automatically added. This variable is ignored if 
              file is provided. Defaults to 'Model'.

        Returns:
            dict: class instance
        """
        return _load(self, file, path, filename)

    def _pix(self, p):
        raise NotImplementedError('No _pix() function defined')

    def _predict_curve(self, args):
        xdata, x = args
        p = np.unravel_index(x, self.shape)
        pix = self._pix(p)
        return pix.predict(xdata)

    def predict(self, xdata: np.ndarray) -> np.ndarray:
        """Predict the data for given x-values.

        Args:
            xdata (array-like): An array with x-values (time points).

        Returns:
            np.ndarray: Array of predicted y-values.
        """

        nx = np.prod(self.shape)
        nt = np.size(xdata)
        if not self.parallel:
            if self.verbose > 0:
                iterator = tqdm(
                    range(nx), 
                    desc='Running predictions for '+self.__class__.__name__)
            else:
                iterator = range(nx)
            ydata = [self._predict_curve((xdata, x)) for x in iterator]
        else:
            args = [(xdata, x) for x in range(nx)]
            pool = multiprocessing.Pool(processes=num_workers)
            ydata = pool.map(self._predict_curve, args)
            pool.close()
            pool.join()
        return np.stack(ydata).reshape(self.shape + (nt,))

    def _train_curve(self, args):
        xdata, ydata, kwargs, x = args
        p = np.unravel_index(x, self.shape)
        # if np.array_equal(p, [2,2]):
        #     print('')
        pix = self._pix(p)
        pix.train(xdata, ydata[x, :], **kwargs)
        if hasattr(pix, 'pcov'):
            sdev = np.sqrt(np.diag(pix.pcov))
        else:
            sdev = np.zeros(len(self.free.keys()))
        for i, par in enumerate(self.free.keys()):
            getattr(self, par)[p] = getattr(pix, par)
            getattr(self, 'sdev_' + par)[p] = sdev[i]
        return pix, p

    def train(self, xdata, ydata: np.ndarray, **kwargs):
        """Train the free parameters

        Args:
            xdata (array-like): Array with x-data (time points)
            ydata (array-like): Array with y-data (signal data)
            kwargs: any keyword parameters accepted by 
              `scipy.optimize.curve_fit`.

        Returns:
            Model: A reference to the model instance.
        """
        nx = np.prod(self.shape)
        nt = ydata.shape[-1]
        if not self.parallel:
            if self.verbose==1:
                iterator = tqdm(range(nx), desc='Training ' +
                                self.__class__.__name__)
            else:
                iterator = range(nx)
            
            for x in iterator:
                args_x = (xdata, ydata.reshape((nx, nt)), kwargs, x)
                self._train_curve(args_x)
        else:
            args = [(xdata, ydata.reshape((nx, nt)), kwargs, x)
                    for x in range(nx)]
            pool = multiprocessing.Pool(processes=num_workers)
            pool.map(self._train_curve, args)
            pool.close()
            pool.join()
        return self

    def cost(self, xdata, ydata, metric='NRMS') -> float:
        """Goodness-of-fit value.

        Args:
            xdata (array-like): Array with x-data (time points).
            ydata (array-like): Array with y-data (signal values)
            metric (str, optional): Which metric to use - options are: 
                **RMS** (Root-mean-square);
                **NRMS** (Normalized root-mean-square); 
                **AIC** (Akaike information criterion); 
                **cAIC** (Corrected Akaike information criterion for small 
                  models);
                **BIC** (Baysian information criterion). Defaults to 'NRMS'.

        Returns:
            np.ndarray: goodness of fit in each element of the data array. 
        """
        return _cost(self, xdata, ydata, metric)

    def export_params(self) -> dict:
        """Model parameters with descriptions.

        Returns:
            dict: Dictionary with one item for each model parameter. The key 
              is the parameter symbol (short name), and the value is a 
              4-element list with [parameter name, value, unit, sdev].
        """
        pars = {}
        for p in self.free.keys():
            pars[p] = [
                p + ' name',
                getattr(self, p),
                p + ' unit',
                getattr(self, 'sdev_' + p),
            ]
        return pars

    def _add_sdev(self, pars):
        for par in pars:
            if par in self.free:
                sdev = getattr(self, 'sdev_' + par)
            else:
                sdev = None
            pars[par].append(sdev)
        return pars

    def _set_defaults(self, free=None, **params):
        for k, v in params.items():
            if hasattr(self, k):
                val = getattr(self, k)
                if isinstance(val, np.ndarray):
                    if isinstance(v, np.ndarray):
                        if v.shape != self.shape:
                            raise ValueError("""Parameter """ + str(k) + """
                                 does not have the correct shape. """)
                        else:
                            setattr(self, k, v)
                    else:
                        setattr(self, k, np.full(self.shape, v))
                elif val is None:
                    setattr(self, k, v)
                else:
                    if isinstance(v, np.ndarray):
                        msg = str(k) + " is not a pixel-based quantity. \n"
                        msg += "Please provide a scalar initial value instead."
                        raise ValueError(msg)
                    else:
                        setattr(self, k, v)
            else:
                raise ValueError(
                    str(k) + ' is not a valid parameter for this model.')
        if free is not None:
            self.set_free(**free)

# TODO: self.free - > self._free and include self.free()
class Model:
    # Abstract base class for end-to-end models.

    def __init__(self):
        self.free = {}
        self.pcov = None

    def _par_values(self, export=False):
        # Abstract method - needs to be overridden
        return {par: getattr(self, par) for par in self.__dict__ 
                if par not in ['free', 'pcov']}

    def _model_pars(self):
        # Abstract method - needs to be overridden
        return list(self._par_values().keys())
    
    def _params(self):
        # Abstract method - needs to be overridden
        return

    def _set_defaults(self, free=None, **params):

        # Model parameters
        model_pars = self._model_pars()
        pars = self._params()
        for p in model_pars:
            setattr(self, p, pars[p]['init'])
        free_pars = [p for p in model_pars if pars[p]['default_free']]
        self.free = {p: pars[p]['bounds'] for p in free_pars}

        # Override defaults
        for k, v in params.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                if k in self._params():
                    setattr(self, k, v)
                else:
                    raise ValueError(
                        str(k) + ' is not a valid parameter for this model.')
            
        # Set free
        if free is not None:
            self.free = {}
            self.set_free(**free)

    def export_params(self, type='dict') -> dict:
        """Return model parameters with their descriptions

        Args:
            type (str, optional): Type of output. If 'dict', a dictionary is 
              returned. If 'list', a list is returned. Defaults to 'dict'.

        Returns:
            dict: Dictionary with one item for each model parameter. The key 
            is the short parameter name, and the value is a 
            4-element list with [long parameter name, value, unit, sdev].

            or:

            list: List with one element for each model parameter. Each 
            element is a list with [short parameter name, 
            long parameter name, value, unit, sdev].
        """
        # Short name, full name, value, units.
        pars = self._par_values(export=True)
        params = self._params()
        pars = {p: [params[p]['name'], pars[p], params[p]['unit']]
                for p in pars}
        pars = self._add_sdev(pars)
        if type == 'dict':
            return pars
        elif type == 'list':
            return [[k] + v for k, v in pars.items()]
        else:
            raise ValueError('Type must be either "dict" or "list".')

    def save(self, file=None, path=None, filename='Model'):
        """Save the current state of the model

        Args:
            file (str, optional): complete path of the file. If this is not 
              provided, a file is constructure from path and filename 
              variables. Defaults to None.
            path (str, optional): path to store the state if file is not 
              provided. Thos variable is ignored if file is provided. 
              Defaults to current working directory.
            filename (str, optional): filename to store the state if file is 
              not provided. If no extension is included, the extension '.pkl' 
              is automatically added. This variable is ignored if file is 
              provided. Defaults to 'Model'.

        Returns:
            dict: class instance
        """
        return _save(self, file, path, filename)

    def load(self, file=None, path=None, filename='Model'):
        """Load the saved state of the model

        Args:
            file (str, optional): complete path of the file. If this is not 
              provided, a file is constructure from path and filename 
              variables. Defaults to None.
            path (str, optional): path to store the state if file is not 
              provided. Thos variable is ignored if file is provided. 
              Defaults to current working directory.
            filename (str, optional): filename to store the state if file is 
              not provided. If no extension is included, the extension 
              '.pkl' is automatically added. This variable is ignored if file 
              is provided. Defaults to 'Model'.

        Returns:
            dict: class instance
        """
        return _load(self, file, path, filename)

    def predict(self, xdata):
        """Predict the data at given xdata

        Args:
            xdata (array-like): Either an array with x-values (time points) 
              or a tuple with multiple such arrays

        Returns:
            tuple or array-like: Either an array of predicted y-values (if 
              xdata is an array) or a tuple of such arrays (if xdata is a 
              tuple).
        """
        raise NotImplementedError('No predict function provided')

    def train(self, xdata, ydata, **kwargs):
        """Train the free parameters

        Args:
            xdata (array-like): Array with x-data (time points)
            ydata (array-like): Array with y-data (signal data)
            kwargs: any keyword parameters accepted by 
              `scipy.optimize.curve_fit`.

        Returns:
            Model: A reference to the model instance.
        """
        return train(self, xdata, ydata, **kwargs)

    def plot(self, xdata, ydata, xlim=None, ref=None, fname=None, show=True):
        """Plot the model fit against data

        Args:
            xdata (array-like): Array with x-data (time points)
            ydata (array-like): Array with y-data (signal data)
            xlim (array_like, optional): 2-element array with lower and upper 
              boundaries of the x-axis. Defaults to None.
            ref (tuple, optional): Tuple of optional test data in the form 
              (x,y), where x is an array with x-values and y is an array with 
              y-values. Defaults to None.
            fname (path, optional): Filepath to save the image. If no value 
              is provided, the image is not saved. Defaults to None.
            show (bool, optional): If True, the plot is shown. Defaults 
              to True.
        """
        raise NotImplementedError(
            'No plot function implemented for model ' + self.__class__.__name__)

    def cost(self, xdata, ydata, metric='NRMS') -> float:
        """Return the goodness-of-fit

        Args:
            xdata (array-like): Array with x-data (time points).
            ydata (array-like): Array with y-data (signal values)
            metric (str, optional): Which metric to use (see notes for 
              possible values). Defaults to 'NRMS'.

        Returns:
            float: goodness of fit.

        Notes:

            Available options are: 
            
            - 'RMS': Root-mean-square.
            - 'NRMS': Normalized root-mean-square. 
            - 'AIC': Akaike information criterion. 
            - 'cAIC': Corrected Akaike information criterion for small 
              models.
            - 'BIC': Baysian information criterion.

        """
        return _cost(self, xdata, ydata, metric)
    

    def print_params(self, round_to=None):
        """Print the model parameters and their uncertainties

        Args:
            round_to (int, optional): Round to how many digits. If this is 
              not provided, the values are not rounded. Defaults to None.
        """
        pars = self.export_params()
        print('')
        print('--------------------------------')
        print('Free parameters with their stdev')
        print('--------------------------------')
        print('')
        for par in self.free:
            p = pars[par]
            if round_to is None:
                v = p[1]
                verr = p[3]
            else:
                v = round(p[1], round_to)
                verr = round(p[3], round_to)
            print(p[0] + ' ('+par+'): ' + str(v) +
                  ' (' + str(verr) + ') ' + p[2])
        print('')
        print('----------------------------')
        print('Fixed and derived parameters')
        print('----------------------------')
        print('')
        for par in pars:
            if par not in self.free:
                p = pars[par]
                if round_to is None:
                    v = p[1]
                else:
                    v = np.round(p[1], round_to)
                print(p[0] + ' ('+par+'): ' + str(v) + ' ' + p[2])

    def params(self, *args, round_to=None):
        """Return the parameter values

        Args:
            args (tuple): parameters to get
            round_to (int, optional): Round to how many digits. If this is 
            not provided, the values are not rounded. Defaults to None.

        Returns:
            list or float: values of parameter values, or a scalar value if 
            only one parameter is required.
        """
        return params(self, *args, round_to=round_to)


    def set_free(self, pop=None, **kwargs):
        """Set the free model parameters.

        Args:
            pop (str or list): a single variable or a list of variables to 
              remove from the list of free parameters. 

        Raises:
            ValueError: if the pop argument contains a parameter that is not 
              in the list of free parameters.
            ValueError: If the parameter is not a model parameter, or bounds 
              are not properly formatted.
        """
        set_free(self, pop=pop, **kwargs)

    def _getflat(self, attr: np.ndarray = None) -> np.ndarray:
        if attr is None:
            attr = self.free.keys()
        vals = []
        for a in attr:
            v = getattr(self, a)
            vals = np.append(vals, np.ravel(v))
        return vals

    def _setflat(self, vals: np.ndarray, pcov: np.ndarray = None, attr: np.ndarray = None):
        if attr is None:
            attr = self.free.keys()
        if pcov is not None:
            perr = np.sqrt(np.diag(pcov))
        else:
            perr = np.zeros(np.size(vals))
        i = 0
        for p in attr:
            v = getattr(self, p)
            if np.isscalar(v):
                v = vals[i]
                e = perr[i]
                i += 1
            else: # Needs a better solution
                n = np.size(v)
                v = np.reshape(vals[i:i+n], np.shape(v))
                e = np.reshape(perr[i:i+n], np.shape(v))
                i += n
            setattr(self, p, v)
            setattr(self, p + '_sdev', e)


    def _add_sdev(self, pars):
        for par in pars:
            pars[par].append(0)
        # if not hasattr(self, 'pcov'):
        #     perr = np.zeros(len(self.free.keys()))
        # elif self.pcov is None:
        #     perr = np.zeros(len(self.free.keys()))
        # else:
        #     perr = np.sqrt(np.diag(self.pcov))
        # for i, par in enumerate(self.free.keys()):
        #     if par in pars:
        #         pars[par][-1] = perr[i]
        for par in self.free.keys():
            if par in pars:
                if hasattr(self, par + '_sdev'):
                    pars[par][-1] = getattr(self, par + '_sdev')
        return pars
    
    # def _x_scale(self):
    #     n = len(self.free)
    #     xscale = np.ones(n)
    #     for p in range(n):
    #         if np.isscalar(self.free[0]):
    #             lb = self.free[0]
    #         else:
    #             lb = self.free[0][p]
    #         if np.isscalar(self.free[1]):
    #             ub = self.free[1]
    #         else:
    #             ub = self.free[1][p]
    #         if (not np.isinf(lb)) and (not np.isinf(ub)):
    #             xscale[p] = ub-lb
    #     return xscale
    

def params(self, *args, round_to=None):
    p = self._par_values()
    if args == ():
        args = p.keys()
    pars = []
    for a in args:
        if a in p:
            v = p[a]
        else:
            if hasattr(self, a):
                v = getattr(self, a)
            else:
                raise ValueError(
                    a + ' is not a model parameter, and cannot be '
                    + 'derived from the model parameters.'
                )
        if round_to is not None:
            v = round(v, round_to)
        pars.append(v)
    if len(pars) == 1:
        return pars[0]
    else:
        return pars
    

def set_free(self, pop=None, **kwargs):
    if pop is not None:
        if np.isscalar(pop):
            if pop in self.free:
                self.free.pop(pop)
            else:
                raise ValueError(
                    pop + ' is not currently a free parameter, so cannot be '
                    'removed from the list.')
        else:
            for par in pop:
                if par in self.free:
                    self.free.pop(par)
                else:
                    raise ValueError(
                        par + ' is not currently a free parameter, so cannot '
                        'be removed from the list.')
    for k, v in kwargs.items():
        if k in self.__dict__:
            if np.size(v) == 2:
                if v[0] < v[1]:
                    if (v[0] <= np.min(getattr(self, k))) and (v[1] >= np.max(getattr(self, k))):
                        self.free[k] = v
                    else:
                        raise ValueError(
                            'Cannot set parameter bounds for ' + str(k) + '. '
                            'The current value'
                            ' ' + str(getattr(self, k)) + ' is outside of '
                            'the bounds. ')
                else:
                    raise ValueError(
                        str(v) + ' is not a proper parameter ''bound. The '
                        'first element must be smaller than the second.')
            else:
                raise ValueError(
                    str(v) + ' is not a proper parameter bound. Bounds must be '
                    'lists or arrays with 2 elements.')
        else:
            raise ValueError(str(k) + ' is not a model parameter.')


def _save(model, file=None, path=None, filename='Model'):
    if file is None:
        if path is None:
            path = os.getcwd()
        if filename.split('.')[-1] != 'pkl':
            filename += '.pkl'
        file = os.path.join(path, filename)
    elif file.split('.')[-1] != 'pkl':
        file += '.pkl'
    with open(file, 'wb') as f:
        pickle.dump(model, f, protocol=pickle.HIGHEST_PROTOCOL)
    f.close
    return model


def _load(model, file=None, path=None, filename='Model'):
    if file is None:
        if path is None:
            path = os.getcwd()
        if filename.split('.')[-1] != 'pkl':
            filename += '.pkl'
        file = os.path.join(path, filename)
    elif file.split('.')[-1] != 'pkl':
        file += '.pkl'
    with open(file, 'rb') as f:
        saved = pickle.load(f)
    model.__dict__.update(saved.__dict__)
    f.close
    return model


def train(model: Model, xdata, ydata, **kwargs):

    if isinstance(ydata, tuple):
        y = np.concatenate(ydata)
    else:
        y = ydata

    def fit_func(_, *p):
        model._setflat(p)
        yp = model.predict(xdata)
        if isinstance(yp, tuple):
            return np.concatenate(yp)
        else:
            return yp

    p0 = model._getflat()
    bounds = [
        [par[0] for par in model.free.values()],
        [par[1] for par in model.free.values()],
    ]
    try:
        pars, model.pcov = curve_fit(
            fit_func, None, y, p0,
            bounds=bounds,  # x_scale=model._x_scale(),
            **kwargs)
    except RuntimeError as e:
        msg = 'Runtime error in curve_fit -- \n'
        msg += str(e) + ' Returning initial values.'
        warnings.warn(msg)
        pars = p0
        model.pcov = np.zeros((np.size(p0), np.size(p0)))
    except ValueError as e:
        msg = 'Value error in curve_fit -- \n'
        msg += str(e) + ' Returning initial values.'
        warnings.warn(msg)
        pars = p0
        model.pcov = np.zeros((np.size(p0), np.size(p0)))

    # Note pcov does not have to be an attribuite
    model._setflat(pars, pcov=model.pcov)

    return model


def _cost(model, xdata, ydata, metric='NRMS') -> float:

    # Predict data at all xvalues
    y = model.predict(xdata)
    if isinstance(ydata, tuple):
        y = np.concatenate(y)
        ydata = np.concatenate(ydata)

    # Calclulate the loss at the fitted values
    if metric == 'RMS':
        loss = np.linalg.norm(y - ydata, axis=-1)
    elif metric == 'NRMS':
        ynorm = np.linalg.norm(ydata, axis=-1)
        yerr = np.linalg.norm(y - ydata, axis=-1)
        with np.errstate(divide='ignore', invalid='ignore'):
            loss = 100*yerr/ynorm
    elif metric == 'AIC':
        rss = np.sum((y-ydata)**2, axis=-1)
        n = ydata.shape[-1]
        k = len(model.free.keys())
        with np.errstate(divide='ignore'):
            loss = k*2 + n*np.log(rss/n)
    elif metric == 'cAIC':
        rss = np.sum((y-ydata)**2)
        n = ydata.shape[-1]
        k = len(model.free.keys())
        with np.errstate(divide='ignore'):
            loss = k*2 + n*np.log(rss/n) + 2*k*(k+1)/(n-k-1)
    elif metric == 'BIC':
        rss = np.sum((y-ydata)**2, axis=-1)
        n = ydata.shape[-1]
        k = len(model.free.keys())
        with np.errstate(divide='ignore'):
            loss = k*np.log(n) + n*np.log(rss/n)
    return loss



import numpy as np
import pydmr
import dcmri as dc


def test_fetch():
    data = dc.fetch('tristan_rats_healthy_six_drugs')
    dmr = pydmr.read(data)
    assert 'FA' in dmr['data']


def test_ca_injection():

    weight = 70
    conc = 0.5
    t0 = 5
    dose = 0.2
    rate = 3
    dt = 0.1

    t = np.arange(0, 20, dt)
    j = dc.ca_injection(t, weight, conc, dose, rate, t0)

    assert np.around(np.sum(j)*dt) == np.around(weight*dose*conc)

    # Test exceptions
    try:
        j = dc.ca_injection(t, 0*weight, conc, dose, rate, t0)
    except:
        assert True
    else:
        assert False

    try:
        j = dc.ca_injection(t, weight, conc, 0.01*dose, rate, t0)
    except:
        assert True
    else:
        assert False


def test_ca_conc():

    assert dc.ca_conc('gadoxetate') == 0.25
    assert dc.ca_conc('gadobutrol') == 1.0
    assert dc.ca_conc('gadobenate') == 0.5
    try:
        dc.ca_conc('MyAgent') == 1.0
    except:
        assert True
    else:
        assert False

def test_ca_std_dose():

    assert dc.ca_std_dose('gadoxetate') == 0.1
    assert dc.ca_std_dose('gadobutrol') == 0.1
    assert dc.ca_std_dose('gadopiclenol') == 0.1
    assert dc.ca_std_dose('gadoterate') == 0.2
    try:
        dc.ca_std_dose('myagent') == 1.0
    except:
        assert True
    else:
        assert False
        
def test_relaxivity():
    assert dc.relaxivity(4.7, 'blood', 'gadobutrol') == 1000*4.7
    assert dc.relaxivity(4.7, 'plasma', 'gadobutrol') == 1000*4.7
    try:
        dc.relaxivity(4.7, 'water', 'gadobutrol')
    except:
        assert True
    else:
        assert False
    assert dc.relaxivity(3.0, 'hepatocytes', 'gadoxetate') == 9800
    assert dc.relaxivity(3.0, 'hepatocytes', 'gadodiamide') == 4000

def test_T1():
    assert dc.T1(4.7, 'liver') == 1/1.281
    try:
        dc.T1(4.7, 'hair')
    except:
        assert True
    else:
        assert False

def test_T2():
    assert dc.T2(1.5, 'csf') == 1.99
    try:
        dc.T2(4.7, 'hair')
    except:
        assert True
    else:
        assert False

def test_PD():
    assert dc.PD('csf') == 0.98
    try:
        dc.PD('hair')
    except:
        assert True
    else:
        assert False

def test_perfusion():
    assert dc.perfusion('Fb', 'csf') == 0.0
    assert dc.perfusion('vb', 'csf') == 0.0
    assert dc.perfusion('PS', 'csf') == 0.0
    assert dc.perfusion('vi', 'csf') == 0.0
    try:
        dc.perfusion('Fb', 'hair')
    except:
        assert True
    else:
        assert False

def test_shepp_logan():

    n=64
    roi = dc.shepp_logan(n=n)
    im = dc.shepp_logan('T1', 'T2', 'PD', 'Fb', 'vb', 'PS', 'vi', n=n)

    vals = im['Fb'][roi['CSF left']==1]
    assert 0 == np.amin(vals)
    assert 0 == np.amax(vals)
    vals = im['vb'][roi['CSF left']==1]
    assert 0 == np.amin(vals)
    assert 0 == np.amax(vals)
    vals = im['PS'][roi['CSF left']==1]
    assert 0 == np.amin(vals)
    assert 0 == np.amax(vals)
    vals = im['vi'][roi['CSF left']==1]
    assert 0 == np.amin(vals)
    assert 0 == np.amax(vals)

    # Special case - 1 parameter - does not return dict
    im = dc.shepp_logan('Fb', n=n)
    vals = im[roi['CSF left']==1]
    assert 0 == np.amin(vals)
    assert 0 == np.amax(vals)




if __name__ == "__main__":

    test_fetch()
    test_ca_injection()
    test_ca_conc()
    test_ca_std_dose()
    test_relaxivity()
    test_T1()
    test_T2()
    test_PD()
    test_perfusion()
    test_shepp_logan()

    print('All lib tests passed!!')



from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np

import dcmri.ui as ui
import dcmri.lib as lib
import dcmri.sig as sig
import dcmri.utils as utils
import dcmri.pk as pk
import dcmri.pk_aorta as pk_aorta
import dcmri.kidney as kidney


class AortaKidneys(ui.Model):
    """Joint model for signals from aorta and both kidneys.

    This model uses a whole body model to simultaneously predict 
    signals in aorta and kidneys (see :ref:`whole-body-tissues`). 

    See Also:
        `Aorta`, `Kidney`

    Args:
        organs (str, optional): Model for the organs in the whole-body 
          model. The options are 'comp' (one compartment) and '2cxm' 
          (two-compartment exchange). Defaults to 'comp'.
        heartlung (str, optional): Model for the heart-lung system in 
          the whole-body model. Options are 'pfcomp' (plug-flow 
          compartment) or 'chain'. Defaults to 'pfcomp'.
        kidneys (str, optional): Model for the kidneys. Options are 
          '2CF' (Two-compartment filtration) and 'HF' (High-flow). 
          Defaults to '2CF'. 
        sequence (str, optional): imaging sequence model. Possible 
          values are 'SS' (steady-state), 'SR' (saturation-recovery), 
          'SSI' (steady state with inflow correction) and 'lin' 
          (linear). Defaults to 'SS'.
        agent (str, optional): Generic name of the contrast agent 
          injected. Defaults to 'gadoterate'.
        params (dict, optional): values for the model parameters,
          specified as keyword parameters. Defaults are used for any 
          that are not provided. See table 
          :ref:`AortaKidneys-defaults` for a list of parameters and 
          their default values.

    Notes:

        In the table below, if **Bounds** is None, the parameter is fixed 
        during training. Otherwise it is allowed to vary between the 
        bounds given.

        .. _AortaKidneys-defaults:
        .. list-table:: AortaKidneys parameters. 
            :widths: 5 10 5 5 5 10
            :header-rows: 1

            * - Parameter
              - Description
              - Value
              - Unit
              - Bounds
              - Usage
            * - **General**
              - 
              - 
              - 
              - 
              - 
            * - dt
              - Prediction time step
              - 0.25
              - sec
              - None
              - Always
            * - tmax
              - Maximum time predicted
              - 120
              - sec
              - None
              - Always
            * - dose_tolerance
              - Stopping criterion whole body model
              - 0.1
              -
              - None
              - Always
            * - t0
              - Baseline duration
              - 0
              - 
              - None
              - Always
            * - field_strength
              - B0-field
              - 3
              - T
              - None
              - Always
            * - **Injection**
              - 
              -
              - 
              - 
              -
            * - weight
              - Subject weight
              - 70
              - kg
              - None
              - Always
            * - dose
              - Contrast agent dose
              - 0.0125
              - mL/kg
              - None
              - Always
            * - rate
              - Contrast agent injection rate
              - 1
              - mL/kg
              - None
              - Always
            * - **Sequence**
              - 
              -
              - 
              - 
              - 
            * - TR
              - Repetition time
              - 0.005
              - sec
              - None
              - sequence in ['SS', 'SSI']
            * - FA
              - Flip angle
              - 15
              - deg
              - None
              - sequence in ['SR', 'SS', 'SSI']
            * - TC
              - Time to k-space center
              - 0.1
              - sec
              - None
              - sequence == 'SR'
            * - TS
              - Sampling duration
              - 0
              - sec
              - None
              - Always
            * - TF
              - Inflow time
              - 0
              - sec
              - None
              - sequence == 'SSI'
            * - **Aorta**
              - 
              -
              - 
              - 
              -
            * - BAT
              - Bolus arrival time
              - 60
              - sec
              - [0, inf]
              - Always
            * - CO
              - Cardiac output
              - 100
              - mL/sec
              - [0, 300]
              - Always
            * - Thl
              - Heart-lung transit time
              - 10
              - sec
              - [0, 30]
              - Always
            * - Dhl
              - Heart-lung dispersion
              - 0.2
              - 
              - [0.05, 0.95]
              - heartlung in ['pfcomp', 'chain']
            * - To
              - Organs transit time
              - 20
              - sec
              - [0, 60]
              - Always
            * - Eo
              - Organs extraction fraction
              - 0.15
              - 
              - [0, 0.5]
              - organs == '2cxm'
            * - Toe
              - Organs extracellular transit time
              - 120
              - sec
              - [0, 800]
              - organs == '2cxm'
            * - Eb
              - Body extraction fraction
              - 0.05
              - 
              - [0.01, 0.15]
              - Always
            * - R10a
              - Arterial precontrast R1
              - 0.7
              - /sec
              - None
              - Always
            * - S0a
              - Arterial signal scale factor
              - 1
              - a.u.
              - None
              - Always
            * - **Kidneys**
              -
              -
              - 
              - 
              -
            * - H
              - Hematocrit
              - 0.45
              - 
              - None
              - Always
            * - RPF
              - Renal plasma flow
              - 20
              - mL/sec
              - [0, 100]
              - Always
            * - DRF
              - Differential renal function
              - 0.5
              - 
              - [0, 1.0]
              - Always
            * - DRPF
              - Differential renal plasma flow
              - 0.5
              - 
              - [0, 1.0]
              - kidneys == '2CF'
            * - FF
              - Filtration fraction
              - 0.1
              - 
              - [0, 0.3]
              - agent in ['gadoxetate', 'gadobenate']
            * - **Left kidney**
              - 
              -
              - 
              - 
              - 
            * - Ta_lk
              - Left kidney arterial delay
              - 0
              - sec
              - [0, 3]
              - Always
            * - vp_lk
              - Left kidney plasma volume
              - 0.15
              - mL/cm3
              - [0, 0.3]
              - Always
            * - Tt_lk
              - Left kidney tubular transit time
              - 120
              - sec
              - [0, inf]
              - Always
            * - R10_lk
              - Left kidney precontrast R1
              - 0.65
              - 1/sec
              - None
              - Always
            * - S0_lk
              - Left kidney signal scale factor
              - 1.0
              - a.u.
              - [0, inf]
              - Always
            * - vol_lk
              - Left kidney volume
              - 150
              - cm3
              - None
              - Always
            * - **Right kidney**
              - 
              -
              - 
              - 
              -
            * - Ta_rk
              - Right kidney arterial delay
              - 0
              - sec
              - [0, 3]
              - Always
            * - vp_rk
              - Right kidney plasma volume
              - 0.15
              - mL/cm3
              - [0, 0.3]
              - Always
            * - Tt_rk
              - Right kidney tubular transit time
              - 120
              - sec
              - [0, inf]
              - Always
            * - R10_rk
              - Right kidney precontrast R1
              - 0.65
              - /sec
              - None
              - Always
            * - S0_rk
              - Right kidney signal scale factor
              - 1.0
              - a.u.
              - [0, inf]
              - Always
            * - vol_rk
              - Right kidney volume
              - 150
              - cm3
              - None
              - Always


    Example:

        Use the model to fit minipig data with inflow correction:

    .. plot::
        :include-source:
        :context: close-figs

        >>> import numpy as np
        >>> import pydmr
        >>> import dcmri as dc

        Read the dataset:

        >>> datafile = dc.fetch('minipig_renal_fibrosis')
        >>> data = pydmr.read(datafile, 'nest')
        >>> rois, pars = data['rois']['Pig']['Test'], data['pars']['Pig']['Test']

        Create an array of time points:

        >>> time = pars['TS'] * np.arange(len(rois['Aorta']))

        Initialize the tissue:

        >>> aorta_kidneys = dc.AortaKidneys(
        ...     sequence='SSI',
        ...     heartlung='chain',
        ...     organs='comp',
        ...     agent="gadoterate",
        ...     dt=0.25,
        ...     field_strength=pars['B0'],
        ...     weight=pars['weight'],
        ...     dose=pars['dose'],
        ...     rate=pars['rate'],
        ...     R10a=1/dc.T1(pars['B0'], 'blood'),
        ...     R10_lk=1/dc.T1(pars['B0'], 'kidney'),
        ...     R10_rk=1/dc.T1(pars['B0'], 'kidney'),
        ...     vol_lk=85,
        ...     vol_rk=85,
        ...     TR=pars['TR'],
        ...     FA=pars['FA'],
        ...     TS=pars['TS'],
        ...     CO=60,   
        ...     t0=15, 
        ... )

        Define time and signal data

        >>> t = (time, time, time)
        >>> signal = (rois['Aorta'], rois['LeftKidney'], rois['RightKidney'])

        Train the system to the data:

        >>> aorta_kidneys.train(t, signal)

        Plot the reconstructed signals and concentrations:

        >>> aorta_kidneys.plot(t, signal)

        Print the model parameters:

        >>> aorta_kidneys.print_params(round_to=4)
        --------------------------------
        Free parameters with their stdev
        --------------------------------
        Bolus arrival time (BAT): 16.7422 (0.2853) sec
        Inflow time (TF): 0.2801 (0.0133) sec
        Cardiac output (CO): 72.762 (12.4426) mL/sec
        Heart-lung mean transit time (Thl): 16.2249 (0.3069) sec
        Organs blood mean transit time (To): 14.3793 (1.2492) sec
        Body extraction fraction (Eb): 0.0751 (0.0071)
        Heart-lung dispersion (Dhl): 0.0795 (0.0041)
        Renal plasma flow (RPF): 3.3489 (0.7204) mL/sec
        Differential renal function (DRF): 0.9085 (0.0212)
        Differential renal plasma flow (DRPF): 0.812 (0.0169)
        Left kidney arterial mean transit time (Ta_lk): 0.6509 (0.2228) sec
        Left kidney plasma volume (vp_lk): 0.099 (0.0186) mL/cm3
        Left kidney tubular mean transit time (Tt_lk): 46.9705 (3.3684) sec
        Right kidney arterial mean transit time (Ta_rk): 1.4206 (0.2023) sec
        Right kidney plasma volume (vp_rk): 0.1294 (0.0175) mL/cm3
        Right kidney tubular mean transit time (Tt_rk): 4497.8301 (39890.3818) sec
        Aorta signal scaling factor (S0a): 4912.776 (254.2363) a.u.
        ----------------------------
        Fixed and derived parameters
        ----------------------------
        Filtration fraction (FF): 0.0812
        Glomerular Filtration Rate (GFR): 0.2719 mL/sec
        Left kidney plasma flow (RPF_lk): 2.7194 mL/sec
        Right kidney plasma flow (RPF_rk): 0.6295 mL/sec
        Left kidney glomerular filtration rate (GFR_lk): 0.247 mL/sec
        Right kidney glomerular filtration rate (GFR_rk): 0.0249 mL/sec
        Left kidney plasma flow (Fp_lk): 0.032 mL/sec/cm3
        Left kidney plasma mean transit time (Tp_lk): 2.838 sec
        Left kidney vascular mean transit time (Tv_lk): 3.0958 sec
        Left kidney tubular flow (Ft_lk): 0.0029 mL/sec/cm3
        Left kidney filtration fraction (FF_lk): 0.0908
        Left kidney extraction fraction (E_lk): 0.0833
        Right kidney plasma flow (Fp_rk): 0.0074 mL/sec/cm3
        Right kidney plasma mean transit time (Tp_rk): 16.8121 sec
        Right kidney vascular mean transit time (Tv_rk): 17.4762 sec
        Right kidney tubular flow (Ft_rk): 0.0003 mL/sec/cm3
        Right kidney filtration fraction (FF_rk): 0.0395
        Right kidney extraction fraction (E_rk): 0.038
    """

    def __init__(
            self, 
            organs='comp', 
            heartlung='pfcomp', 
            kidneys='2CF', 
            sequence='SS', 
            agent='gadoterate',
            **params,
        ):

        # Check configuration
        if organs not in ['comp','2cxm']:
            raise ValueError(
                f"{organs} is not a valid model for the organs. "
                "Current options are 'comp' and '2cxm'."
            )
        if heartlung not in ['pfcomp', 'chain']:
            raise ValueError(
                f"{heartlung} is not a valid heart-lung system. "
                "Current options are 'pfcomp' and 'chain'."
            )
        if kidneys not in ['2CF', 'HF']:
            raise ValueError(
                f"Kinetic model {kidneys} is not available."
            )
        if sequence not in ['SS', 'SR', 'SSI', 'lin']:
            raise ValueError(
                f"Sequence {sequence} is not available."
            )

        # Set configuration
        self.sequence = sequence
        self.organs = organs
        self.heartlung = heartlung
        self.kidneys = kidneys
        self.agent = agent
    
        # Set defaults
        self._set_defaults(**params)
        # For SSI, S0 needs to be free because TF affects the baseline
        if self.sequence == 'SSI':
            self.free['S0a'] = [0, np.inf]

        # Internal flag
        self._predict = None

    def _params(self):
        return PARAMS_AORTA | PARAMS_KIDNEYS | PARAMS_DERIVED
    
    def _model_pars(self):

        # General
        pars = ['dt', 'tmax', 'dose_tolerance', 't0', 'field_strength']

        # Injection
        pars += ['weight', 'dose', 'rate', 'BAT']

        # Sequence
        pars += ['TS']
        if self.sequence == 'SR':
            pars += ['TC', 'FA']
        elif self.sequence=='SS':
            pars += ['TR', 'FA']
        elif self.sequence=='SSI':
            pars += ['TF', 'TR', 'FA']

        # Aorta
        pars += ['CO', 'Thl', 'To', 'Eb', 'R10a', 'S0a']
        if self.heartlung == 'pfcomp':
            pars += ['Dhl']
        elif self.heartlung == 'chain':
            pars += ['Dhl']
        if self.organs=='2cxm':
            pars += ['Toe', 'Eo']

        # Kidneys
        pars += ['RPF', 'DRF', 'H']
        if self.agent in ['gadoxetate', 'gadobenate']:
            pars += ['FF']
        if self.kidneys == '2CF':
            pars += ['DRPF']
        pars += ['Ta_lk', 'vol_lk', 'vp_lk', 'Tt_lk', 'R10_lk']
        pars += ['Ta_rk', 'vol_rk', 'vp_rk', 'Tt_rk', 'R10_rk']

        return pars
    

    def _par_values(self, export=False):

        if export:
            discard = [
                'dt', 'tmax', 't0', 'weight', 'dose', 
                'rate', 'field_strength', 'dose_tolerance', 'R10a', 
                'TS' , 'TC', 'TR', 'FA', 
                'R10_lk', 'R10_rk', 'H', 
                'vol_rk', 'vol_lk',
            ]
            pars = self._par_values()
            return {p: pars[p] for p in pars if p not in discard}

        pars = self._model_pars()
        p = {par: getattr(self, par) for par in pars}

        # Kidneys
        if 'FF' not in p:
            p['FF'] = _div(p['Eb'], 1-p['Eb'])
        if {'RPF', 'FF'}.issubset(p):   
            p['GFR'] =  p['RPF'] * p['FF']
        if {'DRPF', 'RPF'}.issubset(p): 
            p['RPF_lk'] = p['DRPF'] * p['RPF']
            p['RPF_rk'] = (1 - p['DRPF']) * p['RPF']
        if {'DRF', 'GFR'}.issubset(p):
            p['GFR_lk'] = p['DRF'] * p['GFR']
            p['GFR_rk'] = (1 - p['DRF']) * p['GFR']

        # Kidney LK
        if {'RPF_lk', 'vol_lk'}.issubset(p):
            p['Fp_lk'] = _div(p['RPF_lk'], p['vol_lk'])
        if {'RPF_lk', 'GFR_lk', 'vp_lk', 'vol_lk'}.issubset(p):
            p['Tp_lk'] = _div(p['vp_lk'] * p['vol_lk'], p['RPF_lk']+p['GFR_lk'])
        if {'RPF_lk', 'vp_lk', 'vol_lk'}.issubset(p):
            p['Tv_lk'] = _div(p['vp_lk'] * p['vol_lk'], p['RPF_lk'])
        if {'GFR_lk', 'vol_lk'}.issubset(p):
            p['Ft_lk'] = _div(p['GFR_lk'], p['vol_lk'])
        if {'GFR_lk', 'RPF_lk'}.issubset(p):
            p['FF_lk'] = _div(p['GFR_lk'], p['RPF_lk'])
            p['E_lk'] = _div(p['GFR_lk'], p['GFR_lk']+p['RPF_lk'])

        # Kidney RK
        if {'RPF_rk', 'vol_rk'}.issubset(p):
            p['Fp_rk'] = _div(p['RPF_rk'], p['vol_rk'])
        if {'RPF_rk', 'GFR_rk', 'vp_rk', 'vol_rk'}.issubset(p):
            p['Tp_rk'] = _div(p['vp_rk'] * p['vol_rk'], p['RPF_rk']+p['GFR_rk'])
        if {'RPF_rk', 'vp_rk', 'vol_rk'}.issubset(p):
            p['Tv_rk'] = _div(p['vp_rk'] * p['vol_rk'], p['RPF_rk'])
        if {'GFR_rk', 'vol_rk'}.issubset(p):
            p['Ft_rk'] = _div(p['GFR_rk'], p['vol_rk'])
        if {'GFR_rk', 'RPF_rk'}.issubset(p):
            p['FF_rk'] = _div(p['GFR_rk'], p['RPF_rk'])
            p['E_rk'] = _div(p['GFR_rk'], p['GFR_rk']+p['RPF_rk'])

        return p


    def _conc_aorta(self) -> np.ndarray:
        if self.organs == 'comp':
            organs = ['comp', (self.To,)]
        elif self.organs=='2cxm':
            organs = ['2cxm', ([self.To, self.Toe], self.Eo)]
        if self.heartlung == 'comp':
            heartlung = ['comp', (self.Thl,)]
        elif self.heartlung == 'pfcomp':
            heartlung = ['pfcomp', (self.Thl, self.Dhl)]
        elif self.heartlung == 'chain':
            heartlung = ['chain', (self.Thl, self.Dhl)]
        self.t = np.arange(0, self.tmax, self.dt)
        conc = lib.ca_conc(self.agent)
        Ji = lib.ca_injection(
            self.t, self.weight, conc, self.dose, self.rate, self.BAT,
        )
        Jb = pk_aorta.flux_aorta(
            Ji, E=self.Eb, heartlung=heartlung, organs=organs,
            dt=self.dt, tol=self.dose_tolerance,
        )
        self.ca = Jb/self.CO
        return self.t, self.ca

    def _relax_aorta(self):
        t, cb = self._conc_aorta()
        rb = lib.relaxivity(self.field_strength, 'blood', self.agent)
        return t, self.R10a + rb*cb

    def _predict_aorta(self, xdata: np.ndarray) -> np.ndarray:
        tacq = xdata[1]-xdata[0]
        self.tmax = max(xdata)+tacq+self.dt
        if self.TS is not None:
            self.tmax += self.TS
        t, R1a = self._relax_aorta()
        if self.sequence == 'SR':
            signal = sig.signal_free(self.S0a, R1a, self.TC, self.FA)
        elif self.sequence=='SS':
            signal = sig.signal_ss(self.S0a, R1a, self.TR, self.FA)
        elif self.sequence=='SSI':
            signal = sig.signal_spgr(self.S0a, R1a, self.TF, self.TR, self.FA, n0=1)
        elif self.sequence == 'lin':
            signal = sig.signal_lin(self.S0a, R1a)
        return utils.sample(xdata, t, signal, self.TS)


    def _conc_kidneys(self, sum=True):

        if self.agent in ['gadoxetate', 'gadobenate']:
            FF = self.FF
        else:
            FF = self.Eb/(1-self.Eb)

        t = self.t
        ca_lk = pk.flux(
            self.ca, self.Ta_lk, t=self.t, dt=self.dt, model='plug',
        )
        ca_rk = pk.flux(
            self.ca, self.Ta_rk, t=self.t, dt=self.dt, model='plug',
        )
        GFR = FF * self.RPF
        Ft_lk = self.DRF * GFR / self.vol_lk
        Ft_rk = (1-self.DRF) * GFR / self.vol_rk
        if self.kidneys == '2CF':
            Fp_lk = self.DRPF * self.RPF / self.vol_lk
            C_lk = kidney.conc_kidney(
                ca_lk / (1-self.H), 
                Fp_lk, self.vp_lk, Ft_lk, self.Tt_lk,
                t=self.t, dt=self.dt, sum=sum, kinetics='2CF',
            )
            Fp_rk = (1-self.DRPF) * self.RPF / self.vol_rk
            C_rk = kidney.conc_kidney(
                ca_rk / (1-self.H), 
                Fp_rk, self.vp_rk, Ft_rk, self.Tt_rk,
                t=self.t, dt=self.dt, sum=sum, kinetics='2CF',
            )
        if self.kidneys == 'HF':
            C_lk = kidney.conc_kidney(
                ca_lk / (1-self.H), 
                self.vp_lk, Ft_lk, self.Tt_lk,
                t=self.t, dt=self.dt, sum=sum, kinetics='HF',
            )
            C_rk = kidney.conc_kidney(
                ca_rk / (1-self.H), 
                self.vp_rk, Ft_rk, self.Tt_rk,
                t=self.t, dt=self.dt, sum=sum, kinetics='HF',
            )
        return t, C_lk, C_rk


    def _relax_kidneys(self):
        t, Clk, Crk = self._conc_kidneys()
        rb = lib.relaxivity(self.field_strength, 'blood', self.agent)
        return t, self.R10_lk + rb*Clk, self.R10_rk + rb*Crk


    def _predict_kidneys(self, xdata: tuple[np.ndarray, np.ndarray]) -> tuple[np.ndarray, np.ndarray]:
        t, R1_lk, R1_rk = self._relax_kidneys()
        if self.sequence == 'SR':
            signal_lk = sig.signal_spgr(
                self.S0_lk, R1_lk, self.TC, self.TR, self.FA,
            )
            signal_rk = sig.signal_spgr(
                self.S0_rk, R1_rk, self.TC, self.TR, self.FA,
            )
        elif self.sequence in ['SS', 'SSI']:
            signal_lk = sig.signal_ss(self.S0_lk, R1_lk, self.TR, self.FA)
            signal_rk = sig.signal_ss(self.S0_rk, R1_rk, self.TR, self.FA)
        elif self.sequence == 'lin':
            signal_lk = sig.signal_lin(self.S0_lk, R1_lk)
            signal_rk = sig.signal_lin(self.S0_rk, R1_rk)
        return (
            utils.sample(xdata[0], t, signal_lk, self.TS),
            utils.sample(xdata[1], t, signal_rk, self.TS))

    def conc(self, sum=True):
        """Concentrations in aorta and kidney.

        Args:
            sum (bool, optional): If set to true, the kidney 
              concentrations are the sum over all compartments. If 
              set to false, the compartmental concentrations are 
              returned individually. Defaults to True.

        Returns:
            tuple: time points, aorta blood concentrations, left 
              kidney concentrations, right kidney concentrations.
        """
        t, cb = self._conc_aorta()
        t, Clk, Crk = self._conc_kidneys(sum=sum)
        return t, cb, Clk, Crk

    def relax(self):
        """Relaxation rates in aorta and kidney.

        Returns:
            tuple: time points, aorta relaxation rate, left kidney 
              relaxation rate, right kidney relaxation rate.
        """
        t, R1b = self._relax_aorta()
        t, R1_lk, R1_rk = self._relax_kidneys()
        return t, R1b, R1_lk, R1_rk

    def predict(self, xdata: tuple) -> tuple:
        """Predict the data at given xdata

        Args:
            xdata (tuple): Tuple of 3 arrays with time points for 
              aorta, left kidney and right kidney, in that order. 
              The three arrays can all be different in length and value.

        Returns:
            tuple: Tuple of 3 arrays with signals for aorta, left 
              kidney and right kidney, in that order. The three 
              arrays can all be different in length and value but 
              each has to have the same length as its corresponding 
              array of time points.
        """
        # Public interface
        if self._predict is None:
            signala = self._predict_aorta(xdata[0])
            signal_lk, signal_rk = self._predict_kidneys((xdata[1], xdata[2]))
            return signala, signal_lk, signal_rk
        # Private interface with different input & output types
        elif self._predict == 'aorta':
            return self._predict_aorta(xdata)
        elif self._predict == 'kidneys':
            return self._predict_kidneys(xdata)


    def train(self, xdata: tuple, ydata: tuple, **kwargs):
        """Train the free parameters

        Args:
            xdata (tuple): Tuple of 3 arrays with time points for 
              aorta, left kidney and right kidney, in that order. 
              The three arrays can all be different in length and value.
            ydata (tuple): Tuple of 3 arrays with signals for aorta, 
              left kidney and right kidney, in that order. The three 
              arrays can all be different in length and values but 
              each has to have the same length as its corresponding 
              array of time points.
            kwargs: any keyword parameters accepted by `scipy.optimize.curve_fit`.

        Returns:
            Model: A reference to the model instance.
        """
        # Estimate BAT and S0a from data
        if self.sequence == 'SR':
            Srefb = sig.signal_spgr(1, self.R10a, self.TC, self.TR, self.FA)
            Sref_lk = sig.signal_spgr(1, self.R10_lk, self.TC, self.TR, self.FA)
            Sref_rk = sig.signal_spgr(1, self.R10_rk, self.TC, self.TR, self.FA)
        elif self.sequence=='SS':
            Srefb = sig.signal_ss(1, self.R10a, self.TR, self.FA)
            Sref_lk = sig.signal_ss(1, self.R10_lk, self.TR, self.FA)
            Sref_rk = sig.signal_ss(1, self.R10_rk, self.TR, self.FA)
        elif self.sequence=='SSI':
            Srefb = sig.signal_spgr(1, self.R10a, self.TF, self.TR, self.FA)
            Sref_lk = sig.signal_ss(1, self.R10_lk, self.TR, self.FA)
            Sref_rk = sig.signal_ss(1, self.R10_rk, self.TR, self.FA)
        elif self.sequence=='lin':
            Srefb = sig.signal_lin(1, self.R10a)
            Sref_lk = sig.signal_lin(1, self.R10_lk)
            Sref_rk = sig.signal_lin(1, self.R10_rk)
        n0 = max([np.sum(xdata[0] < self.t0), 1])
        self.S0a = np.mean(ydata[0][:n0]) / Srefb
        n0 = max([np.sum(xdata[1] < self.t0), 1])
        self.S0_lk = np.mean(ydata[1][:n0]) / Sref_lk
        n0 = max([np.sum(xdata[2] < self.t0), 1])
        self.S0_rk = np.mean(ydata[2][:n0]) / Sref_rk
        self.BAT = xdata[0][np.argmax(ydata[0])] - (1-self.Dhl)*self.Thl
        self.BAT = max([self.BAT, 0])

        # Copy all free to restor at the end
        free = deepcopy(self.free)

        # Train free aorta parameters on aorta data
        self._predict = 'aorta'
        pars = list(PARAMS_AORTA.keys())
        self.free = {s: free[s] for s in pars if s in free}
        ui.train(self, xdata[0], ydata[0], **kwargs)

        # Train free kidney parameters on kidney data
        self._predict = 'kidneys'
        pars = list(PARAMS_KIDNEYS.keys())
        self.free = {s: free[s] for s in pars if s in free}
        ui.train(self, (xdata[1], xdata[2]), (ydata[1], ydata[2]), **kwargs)

        # Train all parameters on all data
        self._predict = None
        self.free = free
        return ui.train(self, xdata, ydata, **kwargs)

    def plot(self,
             xdata: tuple,
             ydata: tuple,
             xlim=None, ref=None,
             fname=None, show=True):
        """Plot the model fit against data

        Args:
            xdata (tuple): Tuple of 3 arrays with time points for aorta, left kidney and right kidney, in that order. The three arrays can all be different in length and value.
            ydata (tuple): Tuple of 3 arrays with signals for aorta, left kidney and right kidney, in that order. The three arrays can all be different in length and values but each has to have the same length as its corresponding array of time points.
            xlim (array_like, optional): 2-element array with lower and upper boundaries of the x-axis. Defaults to None.
            ref (tuple, optional): Tuple of optional test data in the form (x,y), where x is an array with x-values and y is an array with y-values. Defaults to None.
            fname (path, optional): Filepath to save the image. If no value is provided, the image is not saved. Defaults to None.
            show (bool, optional): If True, the plot is shown. Defaults to True.
        """
        t, cb, Clk, Crk = self.conc(sum=False)
        sig = self.predict((t, t, t))
        fig, ((ax1, ax2), (ax3, ax4), (ax5, ax6)
              ) = plt.subplots(3, 2, figsize=(10, 12))
        fig.subplots_adjust(wspace=0.3)
        _plot_data1scan(t, sig[0], xdata[0], ydata[0],
                        'Aorta', ax1, xlim,
                        color=['lightcoral', 'darkred'],
                        test=None if ref is None else ref[0])
        _plot_data1scan(t, sig[1], xdata[1], ydata[1],
                        'Left kidney', ax3, xlim,
                        color=['cornflowerblue', 'darkblue'],
                        test=None if ref is None else ref[1])
        _plot_data1scan(t, sig[2], xdata[2], ydata[2],
                        'Right kidney', ax5, xlim,
                        color=['cornflowerblue', 'darkblue'],
                        test=None if ref is None else ref[2])
        cb_lk = Clk[0,:] / (self.vp_lk/(1-self.H))
        cb_rk = Crk[0,:] / (self.vp_rk/(1-self.H))
        _plot_conc_aorta(t, cb, cb_lk, cb_rk, ax2, xlim)
        _plot_conc_kidney(t, Clk, 'Left kidney', ax4, xlim)
        _plot_conc_kidney(t, Crk, 'Right kidney', ax6, xlim)
        if fname is not None:
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()

    def cost(self, xdata: tuple, ydata: tuple, metric='NRMS') -> float:
        """Return the goodness-of-fit

        Args:
            xdata (tuple): Tuple of 3 arrays with time points for aorta, left kidney and right kidney, in that order. The three arrays can all be different in length and value.
            ydata (tuple): Tuple of 3 arrays with signals for aorta, left kidney and right kidney, in that order. The three arrays can all be different in length and values but each has to have the same length as its corresponding array of time points.
            metric (str, optional): Which metric to use - options are: 
                **RMS** (Root-mean-square);
                **NRMS** (Normalized root-mean-square); 
                **AIC** (Akaike information criterion); 
                **cAIC** (Corrected Akaike information criterion for small models);
                **BIC** (Baysian information criterion). Defaults to 'NRMS'.

        Returns:
            float: goodness of fit.
        """
        return super().cost(xdata, ydata, metric)



    


# Helper functions for plotting

def _plot_conc_aorta(t, cb, cb_lk, cb_rk, ax, xlim=None):
    if xlim is None:
        xlim = [t[0], t[-1]]
    ax.set(xlabel='Time (min)', ylabel='Blood concentration (mM)',
           xlim=np.array(xlim)/60)
    ax.plot(t/60, 0*t, color='gray')
    ax.plot(t/60, 1000*cb, linestyle='-',
            color='darkred', linewidth=2.0, label='Aorta')
    ax.plot(t/60, 1000*cb_lk, linestyle='--',
            color='lightcoral', linewidth=2.0, label='Left kidney')
    ax.plot(t/60, 1000*cb_rk, linestyle='-.',
            color='lightcoral', linewidth=2.0, label='Right kidney')
    ax.legend()


def _plot_conc_kidney(t, C, kid, ax, xlim=None):
    if xlim is None:
        xlim = [t[0], t[-1]]
    
    ax.set(xlabel='Time (min)', ylabel=f'{kid} concentration (mM)',
           xlim=np.array(xlim)/60)
    ax.plot(t/60, 0*t, color='gray')
    ax.plot(t/60, 1000*C[0, :], linestyle='-',
            color='darkred', linewidth=2.0, label='Blood')
    ax.plot(t/60, 1000*C[1, :], linestyle='-',
            color='darkcyan', linewidth=2.0, label='Tubuli')
    ax.plot(t/60, 1000*(C[0, :]+C[1, :]), linestyle='-',
            color='darkblue', linewidth=2.0, label='Tissue')
    ax.legend()


def _plot_data1scan(t: np.ndarray, sig: np.ndarray,
                    xdata: np.ndarray, ydata: np.ndarray,
                    roi, ax, xlim, color=['black', 'black'],
                    test=None):
    if xlim is None:
        xlim = [t[0], t[-1]]
    ax.set(xlabel='Time (min)', ylabel=f'{roi} signal (a.u.)', xlim=np.array(xlim)/60)
    ax.plot(xdata/60, ydata, marker='o',
            color=color[0], label='Data', linestyle='None')
    ax.plot(t/60, sig, linestyle='-',
            color=color[1], linewidth=3.0, label='Prediction')
    if test is not None:
        ax.plot(np.array(test[0])/60, test[1], color='black',
                marker='D', linestyle='None', label='Test data')
    ax.legend()



PARAMS_AORTA = {
    'dt': {
        'init': 0.25,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Forward model time step',
        'unit': 'sec',
    },
    'tmax': {
        'init': 120,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Maximum acquisition time',
        'unit': 'sec',
    },
    'dose_tolerance': {
        'init': 0.1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Dose tolerance',
        'unit': '',
    },
    't0': {
        'init': 0,
        'default_free': False,
        'bounds': None,
        'name': 'Baseline duration',
        'unit': 'sec',
    },
    'field_strength': {
        'init': 3.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Magnetic field strength',
        'unit': 'T',
    },

    # Injection
    'weight': {
        'init': 70,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Subject weight',
        'unit': 'kg',
    },
    'dose': {
        'init': lib.ca_std_dose('gadoterate'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Contrast agent dose',
        'unit': 'mL/kg',
    },
    'rate': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Contrast agent injection rate',
        'unit': 'mL/sec',
    },

    # Sequence
    'TR': {
        'init': 0.005,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Repetition time',
        'unit': 'sec',
        'pixel_par': False,
    },
    'FA': {
        'init': 15,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Flip angle',
        'unit': 'deg',
        'pixel_par': False,
    },
    'TC': {
        'init': 0.2,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Time to k-space center',
        'unit': 'sec',
        'pixel_par': False,
    },
    'TS': {
        'init': 0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Sampling time',
        'unit': 'sec',
        'pixel_par': False,
    },
    'TF': {
        'init': 0.50,
        'default_free': True,
        'bounds': [0, 2],
        'name': 'Inflow time',
        'unit': 'sec',
        'pixel_par': False,
    },

    # Kinetics
    'BAT': {
        'init': 60,
        'default_free': True,
        'bounds': [0, np.inf],
        'name': 'Bolus arrival time',
        'unit': 'sec',
    },
    'CO': {
        'init': 100,
        'default_free': True,
        'bounds': [0, 300],
        'name': 'Cardiac output',
        'unit': 'mL/sec',
    },
    'Thl': {
        'init': 10,
        'default_free': True,
        'bounds': [0, 30],
        'name': 'Heart-lung mean transit time',
        'unit': 'sec',
    },
    'Dhl': {
        'init': 0.2,
        'default_free': True,
        'bounds': [0.05, 0.95],
        'name': 'Heart-lung dispersion',
        'unit': '',
    },
    'To': {
        'init': 20,
        'default_free': True,
        'bounds': [0, 60],
        'name': 'Organs blood mean transit time',
        'unit': 'sec',
    },
    'Eo': {
        'init': 0.15,
        'default_free': True,
        'bounds': [0, 0.5],
        'name': 'Organs extraction fraction',
        'unit': '',
    },
    'Toe': {
        'init': 120,
        'default_free': True,
        'bounds': [0, 800],
        'name': 'Organs extravascular mean transit time',
        'unit': 'sec',
    },
    'Eb': {
        'init': 0.05,
        'default_free': True,
        'bounds': [0.01, 0.15],
        'name': 'Body extraction fraction',
        'unit': '',
    },


    # Signal
    'R10a': {
        'init': 0.7,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Aorta precontrast R1',
        'unit': 'Hz',
        'pixel_par': False,
    },
    'S0a': {
        'init': 1.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Aorta signal scaling factor',
        'unit': 'a.u.',
        'pixel_par': True,
    },

}


PARAMS_KIDNEYS = {

    # Both kidneys
    'H': {
        'init': 0.45,
        'default_free': False,
        'bounds': [1e-3, 1 - 1e-3],
        'name': 'Tissue Hematocrit',
        'unit': '',
    },
    'RPF': {
        'init': 20,
        'default_free': True,
        'bounds': [0, 100],
        'name': 'Renal plasma flow',
        'unit': 'mL/sec',
    },
    'DRPF': {
        'init': 0.5,
        'default_free': True,
        'bounds': [0, 1],
        'name': 'Differential renal plasma flow',
        'unit': '',
    },
    'DRF': {
        'init': 0.5,
        'default_free': True,
        'bounds': [0, 1.0],
        'name': 'Differential renal function',
        'unit': '',
    },
    'FF': {
        'init': 0.10,
        'default_free': True,
        'bounds': [0.0, 0.5],
        'name': 'Filtration fraction',
        'unit': '',
    },

    # Left kidney

    'Ta_lk': {
        'init': 0,
        'default_free': True,
        'bounds': [0, 3],
        'name': 'Left kidney arterial mean transit time',
        'unit': 'sec',
    },
    'vp_lk': {
        'init': 0.15,
        'default_free': True,
        'bounds': [0, 0.3],
        'name': 'Left kidney plasma volume',
        'unit': 'mL/cm3',
    },
    'Tt_lk': {
        'init': 120,
        'default_free': True,
        'bounds': [0, np.inf],
        'name': 'Left kidney tubular mean transit time',
        'unit': 'sec',
    },
    'R10_lk': {
        'init': 1/lib.T1(3.0, 'kidney'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Left kidney tissue precontrast R1',
        'unit': 'Hz',
    },
    'S0_lk': {
        'init': 1.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Left kidney signal scaling factor',
        'unit': 'a.u.',
    }, 
    'vol_lk': {
        'init': 150,
        'default_free': False,
        'bounds': None,
        'name': 'Left kidney volume',
        'unit': 'mL',
    },
    
    # Right kidney

    'Ta_rk': {
        'init': 0,
        'default_free': True,
        'bounds': [0, 3],
        'name': 'Right kidney arterial mean transit time',
        'unit': 'sec',
    },
    'vp_rk': {
        'init': 0.15,
        'default_free': True,
        'bounds': [0, 0.3],
        'name': 'Right kidney plasma volume',
        'unit': 'mL/cm3',
    },
    'Tt_rk': {
        'init': 120,
        'default_free': True,
        'bounds': [0, np.inf],
        'name': 'Right kidney tubular mean transit time',
        'unit': 'sec',
    },
    'R10_rk': {
        'init': 1/lib.T1(3.0, 'kidney'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Right kidney tissue precontrast R1',
        'unit': 'Hz',
    },
    'S0_rk': {
        'init': 1.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Right kidney signal scaling factor',
        'unit': 'a.u.',
    },
    'vol_rk': {
        'init': None,
        'default_free': False,
        'bounds': None,
        'name': 'Right kidney volume',
        'unit': 'mL',
    },
}

PARAMS_DERIVED = {

    # Derived parameters

    'GFR': {
        'name': 'Glomerular Filtration Rate',
        'unit': 'mL/sec',
    },
    'RPF_lk': {
        'name': 'Left kidney plasma flow',
        'unit': 'mL/sec',
    },
    'GFR_lk': {
        'name': 'Left kidney glomerular filtration rate',
        'unit': 'mL/sec',
    },
    'Fp_lk': {
        'name': 'Left kidney plasma flow',
        'unit': 'mL/sec/cm3',
    },
    'Tp_lk': {
        'name': 'Left kidney plasma mean transit time',
        'unit': 'sec',
    },
    'Tv_lk': {
        'name': 'Left kidney vascular mean transit time',
        'unit': 'sec',
    },
    'Ft_lk': {
        'name': 'Left kidney tubular flow',
        'unit': 'mL/sec/cm3',
    },
    'FF_lk': {
        'name': 'Left kidney filtration fraction',
        'unit': '',
    },
    'E_lk': {
        'name': 'Left kidney extraction fraction',
        'unit': '',
    },
    'RPF_rk': {
        'name': 'Right kidney plasma flow',
        'unit': 'mL/sec',
    },
    'GFR_rk': {
        'name': 'Right kidney glomerular filtration rate',
        'unit': 'mL/sec',
    },
    'Fp_rk': {
        'name': 'Right kidney plasma flow',
        'unit': 'mL/sec/cm3',
    },
    'Tp_rk': {
        'name': 'Right kidney plasma mean transit time',
        'unit': 'sec',
    },
    'Tv_rk': {
        'name': 'Right kidney vascular mean transit time',
        'unit': 'sec',
    },
    'Ft_rk': {
        'name': 'Right kidney tubular flow',
        'unit': 'mL/sec/cm3',
    },
    'FF_rk': {
        'name': 'Right kidney filtration fraction',
        'unit': '',
    },
    'E_rk': {
        'name': 'Right kidney extraction fraction',
        'unit': '',
    },
}


def _div(a, b):
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.where(b == 0, 0, np.divide(a, b))



import numpy as np
import dcmri as dc


# Debugging mode
VERBOSE = 1
SHOW = True

VERBOSE = 0
SHOW = False


def test_ui_aorta():

    truth = {'BAT': 20}
    time, aif, _, _ = dc.fake_tissue(**truth)
    aorta = dc.Aorta(
        dt = 1.5,
        weight = 70,
        agent = 'gadodiamide',
        dose = 0.2,
        rate = 3,
        field_strength = 3.0,
        TR = 0.005,
        FA = 15,
        TS = 1.5,
        R10 = 1/dc.T1(3.0,'blood'), 
        heartlung = 'chain',
    )
    aorta.train(time, aif, xtol=1e-3)
    aorta.plot(time, aif, show=SHOW)
    rec = aorta.export_params()
    rec_aif = aorta.predict(time)
    assert rec['BAT'][1] == aorta.params('BAT')
    assert np.linalg.norm(aif-rec_aif) < 0.1*np.linalg.norm(aif)
    assert np.abs(rec['BAT'][1]-truth['BAT']) < 0.2*truth['BAT']


if __name__ == "__main__":


    test_ui_aorta()

    print('All ui_aorta tests passed!!')



from dcmri import utils
from dcmri.utils import *

from dcmri import lib
from dcmri.lib import *

from dcmri import pk
from dcmri.pk import *

from dcmri import tissue
from dcmri.tissue import *

from dcmri import pk_aorta
from dcmri.pk_aorta import *

from dcmri import pk_lib
from dcmri.pk_lib import *

from dcmri import kidney
from dcmri.kidney import *

from dcmri import liver
from dcmri.liver import *

from dcmri import pk_inv
from dcmri.pk_inv import *

from dcmri import rel
from dcmri.rel import *

from dcmri import sig
from dcmri.sig import *

from dcmri import fake
from dcmri.fake import *

from dcmri import ui
from dcmri.ui import *

from dcmri import ui_tissue
from dcmri.ui_tissue import *

from dcmri import ui_tissue_ls
from dcmri.ui_tissue_ls import *

from dcmri import ui_tissue_ls_array
from dcmri.ui_tissue_ls_array import *

from dcmri import ui_aorta
from dcmri.ui_aorta import *

from dcmri import ui_aorta_kidneys
from dcmri.ui_aorta_kidneys import *

from dcmri import ui_aorta_liver
from dcmri.ui_aorta_liver import *

from dcmri import ui_aorta_portal_liver
from dcmri.ui_aorta_portal_liver import *

from dcmri import ui_kidney
from dcmri.ui_kidney import *

from dcmri import ui_kidney_cortmed
from dcmri.ui_kidney_cortmed import *

from dcmri import ui_liver
from dcmri.ui_liver import *

from dcmri import data
from dcmri.data import *


import numpy as np


def c_lin(R1, r1) -> np.ndarray:
    """Derive concentrations from relaxation rates using a linear relationship.

    Args:
        R1 (float or array-like): Relaxation rates. For a multi-compartmental 
          tissue, R1 is a 2-dimensional array where the first dimension is the 
          number of compartments.
        r1 (float or array-like): relaxivity. For a multi-compartmental 
          tissue, r1 can be a 1-dimensional array with the relaxation rates 
          for each compartment. If it is a scalar, the assumption is that all 
          compartments have the same r1.

    Returns:
        np.ndarray: concentrations in each tissue compartment, in the same 
        shape as R1.
    """
    if R1.ndim == 2:
        c = np.zeros(R1.shape)
        for i in range(R1.ndim):
            if np.isscalar(r1):
                r1i = r1
            else:
                r1i = r1[i]
            c[i, :] = (R1[i, :]-R1[i, 0])/r1i
    else:
        c = (R1-R1[0])/r1
    return c


def relax(c, R10, r1) -> np.ndarray:
    """Derive longitudinal R1 from tissue concentrations assuming a linear 
    relation.

    Args:
        c (array-like): Concentrations, either as a one-dimensionsal 
          array for one-compartment systems, or two-dimensional where the 1st 
          dimension is the number of compartments.
        R10 (array-like or float): Precontrast R1, either a single value for 
          one-compartment tissues or an array with one value for each tissue 
          compartment.
        r1 (float or array-like): relaxivity, either a single value for 
          one-compartment tissues or an array with one value for each tissue 
          compartment.

    Returns:
        np.ndarray: Array with longitudinal relaxivities, same shape as C.
    """

    # One compartment tissues
    if np.isscalar(r1):
        if np.isscalar(R10):
            # c is scalar or 1D
            return R10 + r1*c
        else:
            # concentrations at 1 time point
            if c.shape == R10.shape:
                return R10 + r1*c
            # concentrations at multiple time points
            else:
                return R10[..., np.newaxis] + r1*c

    # n-compartment tissues (compartment is first dimension)
    else:

        r1 = np.array(r1)
        R10 = np.array(R10)
        c = np.array(c)

        n = len(r1)
        R1 = np.zeros(c.shape)
        if R10.ndim == 1:
            if c.shape == R10.shape:
                return R10 + r1*c
            else:
                for i in range(n):
                    R1[i, :] = R10[i] + r1[i] * c[i,:]
                return R1
        else:
            if c.shape == R10.shape:
                for i in range(n):
                    R1[i,...] = R10[i,...] + r1[i] * c[i,...]
                return R1
            else:
                for i in range(n):
                    R1[i,...] = R10[i,...,np.newaxis] + r1[i] * c[i,...]
                return R1




import numpy as np
from scipy.integrate import trapezoid

import dcmri as dc


def test_fake_aif():
    time, aif, _ = dc.fake_aif()
    assert 1800 < trapezoid(aif, time) < 1900
    time, aif, _ = dc.fake_aif(model='SR')
    assert 2680 < trapezoid(aif, time) < 2690

def test_fake_brain():
    time, signal, aif, gt = dc.fake_brain(n=64)
    assert 1800 < trapezoid(aif, time) < 1900
    assert 445 < trapezoid(signal[32,32,:], time) < 460
    time, signal, aif, gt = dc.fake_brain(n=8, model='SR', verbose=1)
    assert 2680 < trapezoid(aif, time) < 2690
    assert 170 < trapezoid(signal[3,3,:], time) < 180

def test_fake_tissue():
    time, aif, roi, gt = dc.fake_tissue()
    assert 1230 < trapezoid(aif, time) < 1240
    assert 1150 < trapezoid(roi, time) < 1170
    time, aif, roi, gt = dc.fake_tissue(model='SR')
    assert 1785 < trapezoid(aif, time) < 1795
    assert 935 < trapezoid(roi, time) < 945

def test_fake_liver():
    time, aif, vif, roi, gt = dc.fake_liver()
    assert 1500 < trapezoid(aif, time) < 1600
    assert 1500 < trapezoid(vif, time) < 1600
    assert 2400 < trapezoid(roi, time) < 2500
    time, aif, vif, roi, gt = dc.fake_liver(sequence='SSI')
    assert 1500 < trapezoid(aif, time) < 1600
    assert 1500 < trapezoid(vif, time) < 1600
    assert 2400 < trapezoid(roi, time) < 2500

def test_fake_tissue2scan():
    time, aif, roi, gt = dc.fake_tissue2scan()
    assert 1230 < trapezoid(aif[0], time[0]) < 1240
    assert 3120 < trapezoid(aif[1], time[1]) < 3145
    assert 1130 < trapezoid(roi[0], time[0]) < 1145
    assert 2930 < trapezoid(roi[1], time[1]) < 2945
    time, aif, roi, gt = dc.fake_tissue2scan(model='SR')
    assert 1785 < trapezoid(aif[0], time[0]) < 1795
    assert 2425 < trapezoid(roi[1], time[1]) < 2435

def test_fake_kidney():
    time, aif, roi, gt = dc.fake_kidney()
    assert 1660 < trapezoid(aif, time) < 1680
    assert 1555 < trapezoid(roi[0], time) < 1565
    assert 900 < trapezoid(roi[1], time) < 915
    time, aif, roi, gt = dc.fake_kidney(model='SS')
    assert 1145 < trapezoid(aif, time) < 1155
    assert 1110 < trapezoid(roi[1], time) < 1125



if __name__ == "__main__":

    test_fake_aif()
    test_fake_tissue()
    test_fake_brain()
    test_fake_tissue2scan()
    test_fake_liver()
    test_fake_kidney()


    print('All fake tests passed!!')